/* i386/jit-icode.h
 * Define the instructions which are present on the i386.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#ifndef __i386_jit_icode_h
#define __i386_jit_icode_h

/*------------------------------------------------------------------------*/
/* Refs and int/long registers are equivalent. */

#define HAVE_ref_eq_int
#undef  HAVE_ref_eq_long

/*------------------------------------------------------------------------*/

#define HAVE_unimplemented		unimplemented
#define HAVE_nop			nop
#define HAVE_prologue			prologue_xxx
#define HAVE_epilogue			epilogue_xxx
#define HAVE_exception_prologue		eprologue_xxx
#define HAVE_spill_int			spill_Rxx
#define HAVE_spill_float		fspill_Rxx
#define HAVE_spill_double		fspilll_Rxx
#define HAVE_reload_int			reload_Rxx
#define HAVE_reload_float		freload_Rxx
#define HAVE_reload_double		freloadl_Rxx
#define HAVE_move_int_const		move_RxC
#define HAVE_move_label_const		move_RxL
#define HAVE_move_int			move_RxR
#define HAVE_move_float_const		fmove_RxC
#define HAVE_move_float			fmove_RxR
#define HAVE_move_double_const		fmovel_RxC
#define HAVE_move_double		fmovel_RxR
#define HAVE_add_int			add_RRR
#define HAVE_adc_int			adc_RRR
#define HAVE_add_float			fadd_RRR
#define HAVE_add_double			faddl_RRR
#define HAVE_sub_int			sub_RRR
#define HAVE_sbc_int			sbc_RRR
#define HAVE_sub_float			fsub_RRR
#define HAVE_sub_double			fsubl_RRR
#define HAVE_mul_int			mul_RRR
#define HAVE_mul_float			fmul_RRR
#define HAVE_mul_double			fmull_RRR
#define HAVE_div_int			div_RRR
#define HAVE_rem_int			rem_RRR
#define HAVE_and_int			and_RRR
#define HAVE_or_int			or_RRR
#define HAVE_xor_int			xor_RRR
#define HAVE_ashr_int			ashr_RRR
#define HAVE_lshr_int			lshr_RRR
#define HAVE_lshl_int			lshl_RRR
#define HAVE_load_int			load_RxR
#define HAVE_load_float			fload_RxR
#define HAVE_load_double		floadl_RxR
#define HAVE_store_int			store_xRR
#define HAVE_store_float		fstore_xRR
#define HAVE_store_double		fstorel_xRR
#define HAVE_cmp_int			cmp_xRR
#define HAVE_cvt_int_float		cvtif_RxR
#define HAVE_cvt_int_double		cvtid_RxR
#define HAVE_cvt_long_float		cvtlf_RxR
#define HAVE_cvt_long_double		cvtld_RxR
#define HAVE_cvt_float_int		cvtfi_RxR
#define HAVE_cvt_double_int		cvtdi_RxR
#define HAVE_cvt_float_long		cvtfl_RxR
#define HAVE_cvt_double_long		cvtdl_RxR
#define HAVE_cvt_float_double		cvtfd_RxR
#define HAVE_cvt_double_float		cvtdf_RxR
#define HAVE_build_key			set_word_xxC
#define HAVE_build_code_ref		set_wordpc_xxC
#define HAVE_set_label			set_label_xxC
#define HAVE_branch			branch_xCC
#define HAVE_branch_indirect		branch_indirect_xRC
#define HAVE_call_ref			call_xCC
#define HAVE_call			call_xRC
#define HAVE_ret			ret_xxx
#define HAVE_pusharg_int		push_xxR
#define HAVE_pusharg_float		fpush_xxR
#define HAVE_pusharg_double		fpushl_xxR
#define HAVE_popargs			popargs_xxC
#define HAVE_returnarg_int		returnarg_RxR
#define HAVE_returnarg_long		returnargl_RxR
#define HAVE_returnarg_float		freturnarg_RxR
#define HAVE_returnarg_double		freturnargl_RxR
#define HAVE_return_int			return_Rxx
#define HAVE_return_long		returnl_Rxx
#define HAVE_return_float		freturn_Rxx
#define HAVE_return_double		freturnl_Rxx
#define HAVE_add_int_const		add_RRC
#define HAVE_adc_int_const		adc_RRC
#define HAVE_sub_int_const		sub_RRC
#define HAVE_load_byte			loadb_RxR
#define HAVE_load_char			loadc_RxR
#define HAVE_load_short			loads_RxR
#define HAVE_store_byte			storeb_xRR
#define HAVE_store_char			stores_xRR
#define HAVE_store_short		stores_xRR
#define HAVE_cmp_int_const		cmp_xRC
#define	HAVE_cmp_load_offset_int	cmp_xRRC
#define HAVE_pusharg_int_const		push_xxC
#define HAVE_lshl_int_const		lshl_RRC
#define HAVE_inval_slot			invalSlot
#define HAVE_start_block		startBlock
#define HAVE_end_block			endBlock
#define HAVE_start_insn			startInsn
#define HAVE_ashr_int_const		ashr_RRC
#define	HAVE_pusharg_int_slot		push_xxS
#define	HAVE_and_int_const		and_RRC
#define	HAVE_neg_int			neg_RxR
#define	HAVE_or_int_const		or_RRC
#define	HAVE_xor_int_const		xor_RRC
#define	HAVE_mul_int_const		mul_RRC
#define	HAVE_lshr_int_const		lshr_RRC

#define	HAVE_store_addr_int		store_xCR
#define HAVE_store_offset_int		store_RRC
#define HAVE_store_offset_byte		storeb_RRC
#define HAVE_store_offset_float		fstore_RRC
#define	HAVE_store_offset_scaled_int	store_RRCR
#define	HAVE_store_offset_scaled_byte	storeb_RRCR
#define	HAVE_store_offset_scaled_byte_const storeb_RRCC
#define	HAVE_store_offset_scaled_short	stores_RRCR
#define	HAVE_store_offset_scaled_char	stores_RRCR

#define	HAVE_load_addr_int		load_RxC
#define HAVE_load_offset_int		load_RRC
#define HAVE_load_offset_float		fload_RRC
#define HAVE_load_offset_byte		loadb_RRC
#define	HAVE_load_offset_scaled_int	load_RRRC
#define	HAVE_load_offset_scaled_byte	loadb_RRRC
#define	HAVE_load_offset_scaled_char	loadc_RRRC

#undef	HAVE_load_addr_float
#undef	HAVE_load_addr_double
#undef	HAVE_store_addr_float
#undef	HAVE_store_addr_double

/*------------------------------------------------------------------------*/

/*
 * These are optional if the architecture supports them.
 */
#undef	HAVE_mul_int_const
#undef	HAVE_swap_int
#undef	HAVE_lshr_int_const

#undef	HAVE_cmpg_float
#undef	HAVE_cmpg_double
#undef	HAVE_cmpl_float
#undef	HAVE_cmpl_double

#undef	HAVE_move_long_const
#undef	HAVE_move_long

#undef	HAVE_add_long
#undef	HAVE_sub_long
#undef	HAVE_mul_long
#undef	HAVE_div_long
#undef	HAVE_rem_long
#undef	HAVE_neg_long
#undef	HAVE_and_long
#undef	HAVE_or_long
#undef	HAVE_xor_long
#undef	HAVE_ashr_long
#undef	HAVE_lshl_long
#undef	HAVE_lshr_long

#undef	HAVE_load_long
#undef	HAVE_load_addr_long
#undef	HAVE_store_long
#undef	HAVE_store_addr_long

#undef	HAVE_pusharg_long

#undef	HAVE_cmp_long

#undef	HAVE_neg_float
#undef	HAVE_neg_double
#undef	HAVE_rem_float
#undef	HAVE_rem_double

#undef	HAVE_cvt_int_long
#undef	HAVE_cvt_int_byte
#undef	HAVE_cvt_int_char
#undef	HAVE_cvt_int_short
#undef	HAVE_cvt_long_int

#undef	HAVE_store_offset_long
#undef	HAVE_store_offset_double
#undef	HAVE_store_offset_char
#undef	HAVE_store_offset_short
#undef	HAVE_store_offset_scaled_long
#undef	HAVE_store_offset_scaled_float
#undef	HAVE_store_offset_scaled_double
#undef	HAVE_store_offset_scaled_short
#undef	HAVE_load_offset_long
#undef	HAVE_load_offset_double
#undef	HAVE_load_offset_char
#undef	HAVE_load_offset_short
#undef	HAVE_load_offset_scaled_long
#undef	HAVE_load_offset_scaled_float
#undef	HAVE_load_offset_scaled_double
#undef	HAVE_load_offset_scaled_short

#endif
