/*
 * i386/fjit.h
 * Common i386 Fast-JIT configuration information.
 *
 * Copyright (c) 1997 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#ifndef __i386_fjit_h
#define __i386_fjit_h

/**/
/* Native function invocation. */
/**/
#define	CALL_KAFFE_FUNCTION_VARARGS(meth, obj, nargs, argptr, retval)	\
	asm volatile ("							\n\
		movl %3,%%eax						\n\
1:		cmpl $0,%%eax						\n\
		je 2f							\n\
		decl %%eax						\n\
		pushl (%4,%%eax,4)					\n\
		jmp 1b							\n\
2:									\n\
		pushl %2						\n\
		call *%1						\n\
		movl %%eax,%0						\n\
		" : "=r" (retval) :					\
	    "r" (meth->ncode), "r" (obj), "r" (nargs), "r" (argptr)	\
	    : "eax", "cc" );						\
	asm volatile ("							\n\
		addl %0,%%esp" : : "r" (4*(nargs+1)) : "cc")

#define	CALL_KAFFE_FUNCTION(meth, obj)					\
	asm("	pusha							\n\
		movl %%esp,%%ebp					\n\
		pushl %1						\n\
		call *%0						\n\
		movl %%ebp,%%esp					\n\
		popa							\n\
		" : : "r" (meth->ncode), "r" (obj))

/**/
/* Exception handling information. */
/**/

extern struct _thread* currentThread;

/* Structure of exception frame on stack */
typedef struct _exceptionFrame {
        uintp	retbp;
        uintp	retpc;
} exceptionFrame;

/* Is this frame valid (ie. is it on the current stack) ? */
#define	FRAMEOKAY(f)							\
	((f)->retbp >= (int)TCTX(currentThread)->stackBase &&	\
	 (f)->retbp < (int)TCTX(currentThread)->stackEnd)

/* Get the next frame in the chain */
#define	NEXTFRAME(f)							\
	(f) = ((exceptionFrame*)(f)->retbp)

/* Extract the PC from the given frame */
#define	PCFRAME(f)							\
	((f)->retpc-1)

/* Extract a local argument from given frame */
#define	FRAMEOBJECT(f)							\
	(*(Hjava_lang_Object**)(f->retbp + 8))

/* Get the first exception frame from a subroutine call */
#define	FIRSTFRAME(f, o)						\
	(f) = *((exceptionFrame*)(((uintp)&(o))-8))

/* Call the relevant exception handler (rewinding the stack as
   necessary). */
#define CALL_KAFFE_EXCEPTION(frame, info, obj)				\
        asm("   movl %2,%%eax						\n\
		movl %0,%%ebp						\n\
		jmp *%1							\n\
		" : : "g" (frame->retbp), "r" (info.handler), "g" (obj) : "eax")

/**/
/* Register management information. */
/**/
#define	MAX_REGISTER_TYPES	2
#define	MAX_INTEGER_REGISTERS	6
#define	MAX_FLOAT_REGISTERS	1
#define	MAX_REGISTERS		(MAX_INTEGER_REGISTERS+MAX_FLOAT_REGISTERS)

#define	INIT_REGISTERS()					\
	ADD_REGISTER(TYPE_INTEGER, 0, 0);  /* eax */		\
	ADD_REGISTER(TYPE_INTEGER, 1, 0);  /* ecx */		\
	ADD_REGISTER(TYPE_INTEGER, 2, 0);  /* edx */		\
	ADD_REGISTER(TYPE_INTEGER, 3, 0);  /* ebx */		\
	ADD_REGISTER(TYPE_INTEGER, 6, 0);  /* esi */		\
	ADD_REGISTER(TYPE_INTEGER, 7, 0);  /* edi */		\
	ADD_REGISTER(TYPE_FLOAT, 0, 0);    /* flt/dbl */

/* Note which types are written back when they are used */
#undef	WRITEBACK_INTEGER
#undef	WRITEBACK_LONG
#define	WRITEBACK_FLOAT
#define	WRITEBACK_DOUBLE

/* Translation between Java types and register types.
 * We don't support longs directly but allow them to be converted into
 * two ints.
 */
#define	TYPE_ILLEGAL		-1
#define	TYPE_INTEGER		 0
#define	TYPE_FLOAT		 1

#define SLOTTYPES {			\
	/* int */	TYPE_INTEGER,	\
	/* long */	TYPE_ILLEGAL,	\
	/* float */	TYPE_FLOAT,	\
	/* double */	TYPE_FLOAT,	\
	/* object */	TYPE_INTEGER,	\
	/* void */	TYPE_ILLEGAL,	\
	/* addr */	TYPE_INTEGER,	\
	/* any */	TYPE_INTEGER,	\
	/* anylong */	TYPE_ILLEGAL,	\
 }

/**/
/* Opcode generation. */
/**/

/* Define if generated code uses two operands rather than one */
#define	TWO_OPERAND

/**/
/* Slot management information. */
/**/

/* Return slots */
#define INIT_RETURNSLOTS()					\
		/* returnInt */					\
	returninfo[0].type = TYPE_INTEGER;			\
	returninfo[0].reg = &registerData[0];   /* eax */	\
		/* returnRef */					\
	returninfo[1].type = TYPE_INTEGER;			\
	returninfo[1].reg = &registerData[0];   /* eax */	\
		/* returnLong */				\
	returninfo[2].type = TYPE_INTEGER;			\
	returninfo[2].reg = &registerData[0];   /* eax */	\
	returninfo[3].type = TYPE_INTEGER;			\
	returninfo[3].reg = &registerData[2];   /* edx */	\
		/* returnFloat */				\
	returninfo[4].type = TYPE_FLOAT;			\
	returninfo[4].reg = &registerData[6];   /* flt/dbl */	\
		/* returnDouble */				\
	returninfo[5].type = TYPE_FLOAT;			\
	returninfo[5].reg = &registerData[6];   /* flt/dbl */	\
	returninfo[6].type = TYPE_ILLEGAL;			\
	returninfo[6].reg = 0

/* Maximum slots allowed in a flow */
#define	MAXFLOWSLOTS	6

/* Size of each slot */
#define	SLOTSIZE	4

/* Generate slot offset for an argument */
#define SLOT2ARGOFFSET(_n)	(8 + SLOTSIZE * ((_n) + isStatic))

/* Generate slot offset for a local (non-argument) */
#define SLOT2LOCALOFFSET(_n)	(-SLOTSIZE * (maxTemp+maxLocal+maxStack - (_n)))

/* We don't have to flush the code out of cache on the i386 */
#define	FLUSH_DCACHE()	/* Do nothing */

#endif
