/*
 * sparc/jit.h
 * Common SPARC JIT configuration information.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#ifndef __sparc_jit_h
#define __sparc_jit_h

/**/
/* Native function invocation. */
/**/
#define	CALL_KAFFE_FUNCTION_VARARGS(meth, obj, nargs, argptr, retval)	\
	asm volatile ("							\n\
		ld [%2],%%o1						\n\
		ld [%2+4],%%o2						\n\
		ld [%2+8],%%o3						\n\
		ld [%2+12],%%o4						\n\
		ld [%2+16],%%o5						\n\
		mov %0,%%o0						\n\
		call %1,0						\n\
		nop							\n\
	" : : "r" (obj), "r" (meth->ncode), "r" (argptr) :		\
		"o0", "o1", "o2", "o3", "o4", "o5" )

#define CALL_KAFFE_FUNCTION(meth, obj)                                  \
	asm("	mov %0,%%o0						\n\
		call %1,0						\n\
		nop							\n\
	" : : "r" (obj), "r" (meth->ncode))

/**/
/* Exception handling information. */
/**/

extern struct _thread* currentThread;

/* Structure of exception frame on stack */
typedef struct _exceptionFrame {
	uintp	lcl[8];
	uintp	arg[6];
        uintp	retbp;
        uintp	retpc;
	uintp	retstruct;
	uintp	callargs[6];
} exceptionFrame;

/* Is this frame valid (ie. is it on the current stack) ? */
#define	FRAMEOKAY(f)							\
	((f)->retbp >= (int)TCTX(currentThread)->stackBase &&	\
	 (f)->retbp < (int)TCTX(currentThread)->stackEnd)

/* Get the next frame in the chain */
#define	NEXTFRAME(f)							\
	(f) = ((exceptionFrame*)(f)->retbp)

/* Extract the PC from the given frame */
#define	PCFRAME(f)							\
	((f)->retpc)

/* Get the first exception frame from a subroutine call */
#define	FIRSTFRAME(f, o)						\
	{								\
		register exceptionFrame* ef asm("%sp");			\
		asm volatile("ta 3");					\
		(f) = *(ef);						\
	}

/* Extract the object argument from given frame */
#define FRAMEOBJECT(f)							\
	(*(Hjava_lang_Object**)(((exceptionFrame*)(f)->retbp)->retbp+68))

/* Call the relevant exception handler (rewinding the stack as
   necessary). */
#define CALL_KAFFE_EXCEPTION(frame, info, obj)				\
	asm volatile("							\n\
		mov %0,%%o1						\n\
		mov %1,%%o2						\n\
		mov %2,%%o3						\n\
		ta 3							\n\
		sub %%sp,64,%%sp					\n\
		mov %%o3,%%fp						\n\
		jmpl %%o2,%%g0						\n\
		restore	%%o1,0,%%o0					\n\
	" : : "r" (obj), "r" (info.handler), "r" (frame->retbp) : "o1", "o2", "o3")

/**/
/* Register management information. */
/**/
#define	MAX_REGISTER_TYPES	2
#define	MAX_INTEGER_REGISTERS	16
#define	MAX_FLOAT_REGISTERS	16
#define	MAX_REGISTERS		(MAX_INTEGER_REGISTERS+MAX_FLOAT_REGISTERS)

#define	INIT_REGISTERS()						\
	ADD_REGISTER(TYPE_INTEGER, 8,  rreserved);	/* o0 */	\
	ADD_REGISTER(TYPE_INTEGER, 9,  rreserved);	/* o1 */	\
	ADD_REGISTER(TYPE_INTEGER, 16, rnocallsave);	/* l0 */	\
	ADD_REGISTER(TYPE_INTEGER, 17, rnocallsave);	/* l1 */	\
	ADD_REGISTER(TYPE_INTEGER, 18, rnocallsave);	/* l2 */	\
	ADD_REGISTER(TYPE_INTEGER, 19, rnocallsave);	/* l3 */	\
	ADD_REGISTER(TYPE_INTEGER, 20, rnocallsave);	/* l4 */	\
	ADD_REGISTER(TYPE_INTEGER, 21, rnocallsave);	/* l5 */	\
	ADD_REGISTER(TYPE_INTEGER, 22, rnocallsave);	/* l6 */	\
	ADD_REGISTER(TYPE_INTEGER, 23, rnocallsave);	/* l7 */	\
	ADD_REGISTER(TYPE_INTEGER, 24, rnocallsave);	/* i0 */	\
	ADD_REGISTER(TYPE_INTEGER, 25, rnocallsave);	/* i1 */	\
	ADD_REGISTER(TYPE_INTEGER, 26, rnocallsave);	/* i2 */	\
	ADD_REGISTER(TYPE_INTEGER, 27, rnocallsave);	/* i3 */	\
	ADD_REGISTER(TYPE_INTEGER, 28, rnocallsave);	/* i4 */	\
	ADD_REGISTER(TYPE_INTEGER, 29, rnocallsave);	/* i5 */	\
	ADD_REGISTER(TYPE_FLOAT, 0, 0);			/* f0 */	\
	ADD_REGISTER(TYPE_FLOAT, 2, 0);			/* f2 */	\
	ADD_REGISTER(TYPE_FLOAT, 4, 0);			/* f4 */	\
	ADD_REGISTER(TYPE_FLOAT, 6, 0);			/* f6 */	\
	ADD_REGISTER(TYPE_FLOAT, 8, 0);			/* f8 */	\
	ADD_REGISTER(TYPE_FLOAT, 10, 0);		/* f10 */	\
	ADD_REGISTER(TYPE_FLOAT, 12, 0);		/* f12 */	\
	ADD_REGISTER(TYPE_FLOAT, 14, 0);		/* f14 */	\
	ADD_REGISTER(TYPE_FLOAT, 16, 0);		/* f16 */	\
	ADD_REGISTER(TYPE_FLOAT, 18, 0);		/* f18 */	\
	ADD_REGISTER(TYPE_FLOAT, 20, 0);		/* f20 */	\
	ADD_REGISTER(TYPE_FLOAT, 22, 0);		/* f22 */	\
	ADD_REGISTER(TYPE_FLOAT, 24, 0);		/* f24 */	\
	ADD_REGISTER(TYPE_FLOAT, 26, 0);		/* f26 */	\
	ADD_REGISTER(TYPE_FLOAT, 28, 0);		/* f28 */	\
	ADD_REGISTER(TYPE_FLOAT, 30, 0);		/* f30 */

/* Note which types are written back when they are used */
#undef	WRITEBACK_INTEGER
#undef	WRITEBACK_LONG
#undef	WRITEBACK_FLOAT
#undef	WRITEBACK_DOUBLE

/* Translation between Java types and register types.
 * We don't support longs directly but allow them to be converted into
 * two ints.
 */
#define TYPE_ILLEGAL            -1
#define TYPE_INTEGER             0
#define TYPE_FLOAT               1

#define	SLOTTYPES {			\
	/* int */	TYPE_INTEGER,	\
	/* long */	TYPE_ILLEGAL,	\
	/* float */	TYPE_FLOAT,	\
	/* double */	TYPE_FLOAT,	\
	/* object */	TYPE_INTEGER,	\
	/* void */	TYPE_ILLEGAL,	\
	/* addr */	TYPE_INTEGER,	\
	/* any */	TYPE_INTEGER,	\
	/* anylong */	TYPE_ILLEGAL,	\
 }

/**/
/* Opcode generation. */
/**/

/**/
/* Slot management information. */
/**/

/* Return slots */
#define INIT_RETURNSLOTS()					\
		/* returnInt */					\
	returninfo[0].type = TYPE_INTEGER;			\
	returninfo[0].reg = &registerData[0];   /* i0 */	\
		/* returnRef */					\
	returninfo[1].type = TYPE_INTEGER;			\
	returninfo[1].reg = &registerData[0];   /* i0 */	\
		/* returnLong */				\
	returninfo[2].type = TYPE_INTEGER;			\
	returninfo[2].reg = &registerData[0];   /* i0 */	\
	returninfo[3].type = TYPE_INTEGER;			\
	returninfo[3].reg = &registerData[2];   /* i1 */	\
		/* returnFloat */				\
	returninfo[4].type = TYPE_FLOAT;			\
	returninfo[4].reg = &registerData[6];   /* f0 */	\
		/* returnDouble */				\
	returninfo[5].type = TYPE_FLOAT;			\
	returninfo[5].reg = &registerData[6];   /* f0+1 */	\
	returninfo[6].type = TYPE_ILLEGAL;			\
	returninfo[6].reg = 0

/* Maximum slots allowed in a flow */
#define	MAXFLOWSLOTS	16

/* Size of each slot */
#define	SLOTSIZE	4

/*
 * A stack frame looks like:
 * 
 * %fp->|                               |
 *      |-------------------------------|
 *      |  Locals, temps, saved floats  |
 *      |-------------------------------|
 *      |  outgoing parameters past 6   |
 *      |-------------------------------|-\
 *      |  6 words for callee to dump   | |
 *      |  register arguments           | |
 *      |-------------------------------|  > minimum stack frame
 *      |  One word struct-ret address  | |
 *      |-------------------------------| |
 *      |  16 words to save IN and      | |
 * %sp->|  LOCAL register on overflow   | |
 *      |-------------------------------|-/
 */

#define	FRAMEALIGN		8
#define	STACKALIGN(v)		(((v) + FRAMEALIGN - 1) & -FRAMEALIGN)

#define	WINDOWSIZE		(16*SLOTSIZE)
#define	INARGSIZE		(6*SLOTSIZE)
#define	WINDOWEXTRA		(5*SLOTSIZE)

/* Generate slot offset for an argument */
#define SLOT2ARGOFFSET(_n) \
	(WINDOWSIZE + SLOTSIZE + SLOTSIZE * ((_n) + isStatic))

/* Generate slot offset for a local (non-argument) */
#define SLOT2LOCALOFFSET(_n) \
	(-(WINDOWEXTRA + SLOTSIZE * ((maxLocal+maxStack+maxTemp) - (_n))))

/* Generate slot offset for an push */
#define SLOT2PUSHOFFSET(_n) \
	(WINDOWSIZE + SLOTSIZE + SLOTSIZE * (_n))

/* On the sparc we need to flush the data cache after generating new code */
#define	FLUSH_DCACHE() \
	{ \
		int i; \
		for (i = 0; i < clen; i += 4) { \
			asm volatile("iflush %0" : : "r" (realcode+i)); \
		} \
	}

#endif
