/* code-analyse.h
 * ...
 *
 * Copyright (c) 1997 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#ifndef __code_analyse_h
#define __code_analyse_h

typedef struct {
	int32			slot;
	int32			value;
	int8			type;
} slotdata;

typedef struct {
	int32			pc;
	int32			value;
	int16			slot;
	int8			type;
} slotalias;

typedef uint8 flowflag;

typedef struct _innerLoop {
        int32                   start;
        int32                   end;
        int32                   in;
        int32                   out;
        struct _innerLoop*      prevflow;
	bool			cancel;
        flowflag                slot[1];
} innerLoop;

typedef struct {
	int16			flags;
	int16			stackPointer;
	int32			fromPc;
	int32			toPc;
	slotdata*		slotAlias;
	innerLoop*		flow;
	int32			pc;
} perPCInfo;

typedef struct codeinfo {
	int32			codelen;
	int32			slotlen;
	perPCInfo*		perPC;
	slotalias*		currentAlias;
} codeinfo;

#define	FLAG_STARTOFBASICBLOCK		0x0001
#define	FLAG_STARTOFEXCEPTION		0x0002
#define	FLAG_STACKPOINTERSET		0x0004
#define	FLAG_CANCELINSN			0x0008
#define	FLAG_NORMALFLOW			0x0010
#define	FLAG_FLOWFORWARD		0x0020
#define	FLAG_FLOWBACK			0x0040
#define	FLAG_FLOWMULTI			0x0080
#define	FLAG_JUMPFORWARD		0x0100
#define	FLAG_JUMPBACK			0x0200
#define	FLAG_JUMPMULTI			0x0400

#define	FLAGS(pc)			codeInfo.perPC[pc].flags
#define	STACKPOINTER(pc)		codeInfo.perPC[pc].stackPointer
#define	FROMPC(pc)			codeInfo.perPC[pc].fromPc
#define	TOPC(pc)			codeInfo.perPC[pc].toPc
#define	FLOW(pc)			codeInfo.perPC[pc].flow
#define	ISLOTS(pc, s)			codeInfo.perPC[pc].slotAlias[s].slot
#define	IVALUES(pc, s)			codeInfo.perPC[pc].slotAlias[s].value
#define	ITYPES(pc, s)			codeInfo.perPC[pc].slotAlias[s].type
#define	OSLOTS(pc, s)			codeInfo.perPC[pc+1].slotAlias[s].slot
#define	OVALUES(pc, s)			codeInfo.perPC[pc+1].slotAlias[s].value
#define	OTYPES(pc, s)			codeInfo.perPC[pc+1].slotAlias[s].type
#define	INSN(pc)			meth->code[(pc)]
#define	INSNLEN(pc)			insnLen[INSN(pc)]
#define	BYTE(pc)			(int8)(meth->code[(pc)+0])
#define	WORD(pc)			(int16)((meth->code[(pc)+0] << 8) + \
						(meth->code[(pc)+1]))
#define	DWORD(pc)			(int32)((meth->code[(pc)+0] << 24) + \
						(meth->code[(pc)+1] << 16) + \
						(meth->code[(pc)+2] << 8) + \
						(meth->code[(pc)+3]))
#define	INCPC(v)			pc += (v)

#define	SNONE				0
#define	SSLOT				1
#define	SCONST				2

#define	SLOTALIAS(s)			codeInfo.currentAlias[(s)].slot
#define	SLOTVALUE(s)			codeInfo.currentAlias[(s)].value
#define	SLOTTYPE(s)			codeInfo.currentAlias[(s)].type
#define	SLOTPC(s)			codeInfo.currentAlias[(s)].pc

#define	SET_STACKALIAS(s, a)		SLOTALIAS(sp+(s)) = (a); \
					SLOTTYPE(sp+(s)) = SSLOT
#define	SET_STACKVALUE(s, a)		SLOTVALUE(sp+(s)) = (a); \
					SLOTTYPE(sp+(s)) = SCONST

#define	SET_STACKPC(s, p)		SLOTPC(sp+(s)) = (p)

#define	SET_STARTOFBASICBLOCK(pc)	FLAGS(pc) |= FLAG_STARTOFBASICBLOCK
#define	SET_STACKPOINTER(pc, sp)	STACKPOINTER(pc) = (sp); \
					FLAGS(pc) |= FLAG_STACKPOINTERSET
#define	SET_FROMPC(pc, from)		FROMPC(pc) = (from)
#define	SET_TOPC(pc, to)		TOPC(pc) = (to)
#define	SET_FLOWMULTI(pc)		if ((FLAGS(pc) & (FLAG_NORMALFLOW|FLAG_FLOWFORWARD|FLAG_FLOWBACK)) != 0) { \
						FLAGS(pc) |= FLAG_FLOWMULTI; \
					}
#define	SET_JUMPMULTI(pc)		if ((FLAGS(pc) & (FLAG_JUMPFORWARD|FLAG_JUMPBACK)) != 0) { \
						FLAGS(pc) |= FLAG_JUMPMULTI; \
					}
#define	SET_NORMALFLOW(pc)		SET_FLOWMULTI(pc); \
					FLAGS(pc) |= FLAG_NORMALFLOW
#define	SET_JUMPFLOW(from, to)		SET_FLOWMULTI(to); \
					SET_JUMPMULTI(from); \
					if (to > from) { \
					  FLAGS(to) |= FLAG_FLOWFORWARD; \
					  FLAGS(from) |= FLAG_JUMPFORWARD; \
					} \
					else { \
					  FLAGS(to) |= FLAG_FLOWBACK; \
					  FLAGS(from) |= FLAG_JUMPBACK; \
					} \
					SET_FROMPC(to, from); \
					SET_TOPC(from, to)
#define	SET_STARTOFEXCEPTION(pc)	FLAGS(pc) |= FLAG_STARTOFEXCEPTION
#define	SET_CANCELINSN(pc)		FLAGS(pc) |= FLAG_CANCELINSN

#define	IS_STARTOFBASICBLOCK(pc)	(FLAGS(pc) & FLAG_STARTOFBASICBLOCK)
#define	IS_STACKPOINTERSET(pc)		(FLAGS(pc) & FLAG_STACKPOINTERSET)
#define	IS_NORMALFLOW(pc)		(FLAGS(pc) & FLAG_NORMALFLOW)
#define	IS_STARTOFEXCEPTION(pc)		(FLAGS(pc) & FLAG_STARTOFEXCEPTION)
#define	IS_CANCELINSN(pc)		(FLAGS(pc) & FLAG_CANCELINSN)
#define	IS_FLOWFORWARD(pc)		(FLAGS(pc) & FLAG_FLOWFORWARD)
#define	IS_FLOWBACK(pc)			(FLAGS(pc) & FLAG_FLOWBACK)
#define	IS_FLOWMULTI(pc)		(FLAGS(pc) & FLAG_FLOWMULTI)
#define	IS_JUMPFORWARD(pc)		(FLAGS(pc) & FLAG_JUMPFORWARD)
#define	IS_JUMPBACK(pc)			(FLAGS(pc) & FLAG_JUMPBACK)
#define	IS_JUMPMULTI(pc)		(FLAGS(pc) & FLAG_JUMPMULTI)

#define	STKPUSH(l)			sp -= (l)
#define	STKPOP(l)			sp += (l)

#define	SLOTIN_TYPE(s)			SLOTTYPE(s)
#define	SLOTIN_CONST(s, i)		slotin_const(pc, s)
#define	SLOTIN_ALIAS(s, i)		slotin_alias(pc, s)

#define	STACKINIT(s, i)			ISLOTS(pc, sp+(s)) = sp+(s)
#define	LOCALINIT(l, i)			ISLOTS(0, l) = (l)

#define	STACKIN(s, i)			SLOTIN_ALIAS(sp+(s), i)
#define	STACKIN_TYPE(s)			SLOTIN_TYPE(sp+(s))
#define	STACKIN_CONST(s, i)		SLOTIN_CONST(sp+(s), i)

#define	STACKIN_SLOT_CONST(s, i)	(STACKIN_TYPE(s) == SCONST ? \
						STACKIN_CONST(s, i) : \
						STACKIN(s, i))

#define	STACKOUT(s, i)			OSLOTS(pc, sp+(s)) = sp+(s); \
					SET_STACKPC((s), pc)
#define	STACKOUT_LOCAL(s, i, l)		OSLOTS(pc, sp+(s)) = sp+(s); \
					SET_STACKALIAS((s), (l)); \
					SET_STACKPC((s), pc)
#define	STACKOUT_CONST(s, i, v)		OSLOTS(pc, sp+(s)) = sp+(s); \
					SET_STACKVALUE((s), (v)); \
					SET_STACKPC((s), pc)

#define	LOCALIN(l, i)			ISLOTS(pc, l) = (l); \
					SET_FLOWSLOT(l)
#define	LOCALOUT(l, i)			OSLOTS(pc, l) = (l); \
					RESET_ONE_ALIAS(l); \
					SET_FLOWSLOT(l)
#define LOCALOUT_STACK(l, i, s)		slotout(pc, sp+(s), l)

#define	RESET_ALIASES()		for (ri = 0; ri < codeInfo.slotlen; ri++) { \
					SLOTALIAS(ri) = ri; \
					SLOTTYPE(ri) = SSLOT; \
					SLOTPC(ri) = 0; \
				}
#define	RESET_ONE_ALIAS(l)	for (ri = 0; ri < codeInfo.slotlen; ri++) { \
					if (SLOTALIAS(ri) == (l)) { \
						SLOTALIAS(ri) = ri; \
						SLOTTYPE(ri) = SSLOT; \
						SLOTPC(ri) = 0; \
					} \
				}

#define	SET_FLOW(pc, f)		(f)->prevflow = flow; \
				flow = (f)
#define	UNSET_FLOW()		nflow = flow->prevflow; \
				if (nflow != 0) { \
				  for (ri = 0; ri < codeInfo.slotlen; ri++) { \
					nflow->slot[ri] |= flow->slot[ri]; \
				  } \
				} \
				if (flow->cancel == false) { \
					FLOW(flow->in) = flow; \
				} \
				flow = nflow
#define	CANCEL_FLOW()		if (flow != 0) { \
					flow->cancel = true; \
				}
#define	SET_FLOWSLOT(s)		if (flow != 0) { \
					flow->slot[s] = 1; \
				}
#define	IS_FLOW(pc)		(FLOW(pc) != 0)
#define	IS_FLOWSLOT(s)		(flow->slot[s] != 0)
					

struct _methods;
void verifyMethod(struct _methods*);

extern codeinfo codeInfo;

#endif
