/* codeproto.h
 * Code prototypes.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#ifndef __codeproto_h
#define __codeproto_h

struct _classes;
struct _label_;

void _start_basic_block(uintp, uintp);
void _end_basic_block(int, int);

void prologue(void);
void epilogue(void);

void move_int(SlotInfo*, SlotInfo*);
void move_long(SlotInfo*, SlotInfo*);
void move_float(SlotInfo*, SlotInfo*);
void move_double(SlotInfo*, SlotInfo*);

void move_int_const(SlotInfo*, jint);
void move_long_const(SlotInfo*, jlong);
void move_float_const(SlotInfo*, jfloat);
void move_double_const(SlotInfo*, jdouble);
void swap_int(SlotInfo*, SlotInfo*);

void add_int(SlotInfo*, SlotInfo*, SlotInfo*);
void add_long(SlotInfo*, SlotInfo*, SlotInfo*);
void add_float(SlotInfo*, SlotInfo*, SlotInfo*);
void add_double(SlotInfo*, SlotInfo*, SlotInfo*);
void sub_int(SlotInfo*, SlotInfo*, SlotInfo*);
void sub_long(SlotInfo*, SlotInfo*, SlotInfo*);
void sub_float(SlotInfo*, SlotInfo*, SlotInfo*);
void sub_double(SlotInfo*, SlotInfo*, SlotInfo*);
void mul_int(SlotInfo*, SlotInfo*, SlotInfo*);
void mul_long(SlotInfo*, SlotInfo*, SlotInfo*);
void mul_float(SlotInfo*, SlotInfo*, SlotInfo*);
void mul_double(SlotInfo*, SlotInfo*, SlotInfo*);
void div_int(SlotInfo*, SlotInfo*, SlotInfo*);
void div_long(SlotInfo*, SlotInfo*, SlotInfo*);
void div_float(SlotInfo*, SlotInfo*, SlotInfo*);
void div_double(SlotInfo*, SlotInfo*, SlotInfo*);
void rem_int(SlotInfo*, SlotInfo*, SlotInfo*);
void rem_long(SlotInfo*, SlotInfo*, SlotInfo*);
void rem_float(SlotInfo*, SlotInfo*, SlotInfo*);
void rem_double(SlotInfo*, SlotInfo*, SlotInfo*);
void neg_int(SlotInfo*, SlotInfo*);
void neg_long(SlotInfo*, SlotInfo*);
void neg_float(SlotInfo*, SlotInfo*);
void neg_double(SlotInfo*, SlotInfo*);
void and_int(SlotInfo*, SlotInfo*, SlotInfo*);
void and_long(SlotInfo*, SlotInfo*, SlotInfo*);
void or_int(SlotInfo*, SlotInfo*, SlotInfo*);
void or_long(SlotInfo*, SlotInfo*, SlotInfo*);
void xor_int(SlotInfo*, SlotInfo*, SlotInfo*);
void xor_long(SlotInfo*, SlotInfo*, SlotInfo*);

void add_int_const(SlotInfo*, SlotInfo*, jint);
void sub_int_const(SlotInfo*, SlotInfo*, jint);

void ashr_int(SlotInfo*, SlotInfo*, SlotInfo*);
void ashr_long(SlotInfo*, SlotInfo*, SlotInfo*);
void lshr_int(SlotInfo*, SlotInfo*, SlotInfo*);
void lshr_long(SlotInfo*, SlotInfo*, SlotInfo*);
void lshl_int(SlotInfo*, SlotInfo*, SlotInfo*);
void lshl_long(SlotInfo*, SlotInfo*, SlotInfo*);

void lshl_int_const(SlotInfo*, SlotInfo*, jint);

void cmp_int(SlotInfo*, SlotInfo*, SlotInfo*);
void cmp_int_const(SlotInfo*, SlotInfo*, jint);
void cmp_load_offset_int(SlotInfo*, SlotInfo*, SlotInfo*, int);
void cmp_long(SlotInfo*, SlotInfo*, SlotInfo*);
void cmpl_float(SlotInfo*, SlotInfo*, SlotInfo*);
void cmpg_float(SlotInfo*, SlotInfo*, SlotInfo*);
void cmpl_double(SlotInfo*, SlotInfo*, SlotInfo*);
void cmpg_double(SlotInfo*, SlotInfo*, SlotInfo*);

void load_int(SlotInfo*, SlotInfo*);
void load_offset_int(SlotInfo*, SlotInfo*, int);
void load_long(SlotInfo*, SlotInfo*);
void load_float(SlotInfo*, SlotInfo*);
void load_double(SlotInfo*, SlotInfo*);
void load_byte(SlotInfo*, SlotInfo*);
void load_char(SlotInfo*, SlotInfo*);
void load_short(SlotInfo*, SlotInfo*);
void load_code_ref(SlotInfo*, SlotInfo*);
void load_key(SlotInfo*, SlotInfo*);
void store_int(SlotInfo*, SlotInfo*);
void store_offset_int(SlotInfo*, int, SlotInfo*);
void store_long(SlotInfo*, SlotInfo*);
void store_float(SlotInfo*, SlotInfo*);
void store_double(SlotInfo*, SlotInfo*);
void store_byte(SlotInfo*, SlotInfo*);
void store_offset_byte(SlotInfo*, int, SlotInfo*);
void store_char(SlotInfo*, SlotInfo*);
void store_short(SlotInfo*, SlotInfo*);

void cvt_int_long(SlotInfo*, SlotInfo*);
void cvt_int_float(SlotInfo*, SlotInfo*);
void cvt_int_double(SlotInfo*, SlotInfo*);
void cvt_int_byte(SlotInfo*, SlotInfo*);
void cvt_int_char(SlotInfo*, SlotInfo*);
void cvt_int_short(SlotInfo*, SlotInfo*);
void cvt_long_int(SlotInfo*, SlotInfo*);
void cvt_long_float(SlotInfo*, SlotInfo*);
void cvt_long_double(SlotInfo*, SlotInfo*);
void cvt_float_int(SlotInfo*, SlotInfo*);
void cvt_float_long(SlotInfo*, SlotInfo*);
void cvt_float_double(SlotInfo*, SlotInfo*);
void cvt_double_int(SlotInfo*, SlotInfo*);
void cvt_double_long(SlotInfo*, SlotInfo*);
void cvt_double_float(SlotInfo*, SlotInfo*);

void pusharg_int(SlotInfo*);
void pusharg_int_const(jint);
void pusharg_long(SlotInfo*);
void pusharg_float(SlotInfo*);
void pusharg_double(SlotInfo*);
void popargs(void);

void call_ref(uintp);
void call(SlotInfo*, SlotInfo*);

void return_int(SlotInfo*);
void return_long(SlotInfo*);
void return_float(SlotInfo*);
void return_double(SlotInfo*);
void ret(void);

void breakpoint(void);

struct _label_* reference_label(int, int);
struct _label_* reference_code_label(int);
struct _label_* reference_table_label(int);
void set_label(int, int);
void move_label_const(SlotInfo*, struct _label_*);
void branch(struct _label_*, int);
void branch_indirect(SlotInfo*);

SlotInfo* stored_code_label(SlotInfo*);
SlotInfo* table_code_label(SlotInfo*);

void build_key(uint8*);

void softcall_lookupmethod(SlotInfo*, Method*, SlotInfo*);
void softcall_badarrayindex(void);
void softcall_nullpointer(void);
void softcall_new(SlotInfo*, struct _classes*);
void softcall_newarray(SlotInfo*, SlotInfo*, int);
void softcall_anewarray(SlotInfo*, SlotInfo*, struct _classes*);
void softcall_multianewarray(SlotInfo*, int, SlotInfo*, struct _classes*);
void softcall_athrow(SlotInfo*);
void softcall_checkcast(SlotInfo*, SlotInfo*, struct _classes*);
void softcall_instanceof(SlotInfo*, SlotInfo*, struct _classes*);
void softcall_monitorenter(SlotInfo*);
void softcall_monitorexit(SlotInfo*);
void softcall_initialise_class(struct _classes*);
void softcall_addtogc(SlotInfo*);

struct _label_* build_code_ref(uint8*, uintp);

#endif
