/* icode-common.h
 * Define the common instruction information.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996-1997
 */

#ifndef __kaffevm_jit_icode_common_h
#define __kaffevm_jit_icode_common_h

/*------------------------------------------------------------------------*/

/*
 * Change various branches into one branch type.
 */
#define	branch_a(l)			branch(l, ba)

#define	cbranch_int_eq(s1, s2, l)	cbranch_int(s1, s2, l, beq)
#define	cbranch_int_ne(s1, s2, l)	cbranch_int(s1, s2, l, bne)
#define	cbranch_int_lt(s1, s2, l)	cbranch_int(s1, s2, l, blt)
#define	cbranch_int_le(s1, s2, l)	cbranch_int(s1, s2, l, ble)
#define	cbranch_int_gt(s1, s2, l)	cbranch_int(s1, s2, l, bgt)
#define	cbranch_int_ge(s1, s2, l)	cbranch_int(s1, s2, l, bge)
#define	cbranch_int_uge(s1, s2, l)	cbranch_int(s1, s2, l, buge)
#define	cbranch_int_ult(s1, s2, l)	cbranch_int(s1, s2, l, bult)
#define	cbranch_load_offset_int_ult(s1, s2, o, l) cbranch_load_offset_int(s1, s2, o, l, bult)

#define	cbranch_int_const_eq(s1, s2, l)	cbranch_int_const(s1, s2, l, beq)
#define	cbranch_int_const_ne(s1, s2, l)	cbranch_int_const(s1, s2, l, bne)
#define	cbranch_int_const_lt(s1, s2, l)	cbranch_int_const(s1, s2, l, blt)
#define	cbranch_int_const_le(s1, s2, l)	cbranch_int_const(s1, s2, l, ble)
#define	cbranch_int_const_gt(s1, s2, l)	cbranch_int_const(s1, s2, l, bgt)
#define	cbranch_int_const_ge(s1, s2, l)	cbranch_int_const(s1, s2, l, bge)
#define	cbranch_int_const_uge(s1, s2, l) cbranch_int_const(s1, s2, l, buge)
#define	cbranch_int_const_ult(s1, s2, l) cbranch_int_const(s1, s2, l, bult)

/*
 * These are for handling unknown types.
 */
#define	move_any			move_int
#define	move_anylong			move_long
#define	pusharg_any			pusharg_int
#define	pusharg_anylong			pusharg_long
#define	load_any			load_int
#define	load_anylong			load_long
#define	store_any			store_int
#define	store_anylong			store_long
#define	load_offset_any			load_offset_int
#define	store_offset_any		store_offset_int

/*
 * These are architecture specific.
 * References may be int's or long's depending on the machine.
 */
#if defined(HAVE_ref_eq_int)

#define	move_ref_const(s, v)		move_int_const(s, (uintp)v)
#define	move_ref			move_int
#define	load_ref			load_int
#define	store_ref			store_int
#define	load_addr_ref			load_addr_int
#define	store_addr_ref			store_addr_int
#define	load_offset_ref			load_offset_int
#define	load_offset_scaled_ref		load_offset_scaled_int
#define	store_offset_ref		store_offset_int
#define	store_offset_scaled_ref		store_offset_scaled_int
#define	pusharg_ref			pusharg_int
#define	pusharg_ref_const(v)		pusharg_int_const((uintp)v);
#define	add_ref_const			add_int_const
#define	add_ref				add_int
#define	cmp_ref_const			cmp_int_const
#define	cmp_ref				cmp_int
#define	return_ref			return_int
#define	returnarg_ref			returnarg_int
#define	cbranch_ref_eq			cbranch_int_eq
#define	cbranch_ref_ne			cbranch_int_ne
#define	cbranch_ref_const_eq		cbranch_int_const_eq
#define	cbranch_ref_const_ne		cbranch_int_const_ne

#elif defined(HAVE_ref_eq_long)

#define	move_ref_const(s, v)		move_long_const(s, (uintp)v)
#define	move_ref			move_long
#define	load_ref			load_long
#define	store_ref			store_long
#define	load_addr_ref			load_addr_long
#define	store_addr_ref			store_addr_long
#define	load_offset_ref			load_offset_long
#define	load_offset_scaled_ref		load_offset_scaled_long
#define	store_offset_ref		store_offset_long
#define	store_offset_scaled_ref		store_offset_scaled_long
#define	pusharg_ref			pusharg_long
#define	pusharg_ref_const(v)		pusharg_long_const((uintp)v);
#define	add_ref_const			add_long_const
#define	add_ref				add_long
#define	cmp_ref_const			cmp_long_const
#define	cmp_ref				cmp_long
#define	return_ref			return_long
#define	returnarg_ref			returnarg_long
#define	cbranch_ref_eq			cbranch_long_eq
#define	cbranch_ref_ne			cbranch_long_ne
#define	cbranch_ref_const_eq		cbranch_long_const_eq
#define	cbranch_ref_const_ne		cbranch_long_const_ne

#endif

/* -------------------------------------------------------------------- */
/* Branches */

#define ba                      0       /* Always */
#define beq                     1       /* Equal */
#define blt                     2       /* Less than */
#define ble                     3       /* Less than or equal */
#define bgt                     4       /* Greater than */
#define bge                     5       /* Greater than or equal */
#define bne                     6       /* Not equal */
#define bn                      7       /* Never */
#define bindirect               8       /* Indirect */
#define buge                    9       /* Unsigned greater than or equal */
#define bult                   10       /* Unsigned less than */

/*------------------------------------------------------------------------*/
/* Various standard functions. */

#define start_instruction()	startInsn(pc)
#define start_function()	prologue()
#define start_basic_block()	startBlock()
#define start_exception_block()	_start_exception_block(&localinfo[stackno])
#define end_basic_block()	endBlock()
#define end_function()		epilogue()
#define monitor_enter()		mon_enter(meth, local(0))
#define monitor_exit()		mon_exit(meth, local(0))
#define	prepare_branch()	end_basic_block()
#define prepare_function_call()	prepareFunctionCall()
#define fixup_function_call()	fixupFunctionCall()
#define	reload_inner()		/* Does nothing */

/*------------------------------------------------------------------------*/

#define	CALL_GEN_FUNC(_f)	issuenr++; _CALL_GEN_FUNC(_f)
#define	_CALL_GEN_FUNC(_f)	HAVE_##_f

#endif
