/* icode.c
 * Define the instructions.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "slots.h"
#include "labels.h"
#include "icode.h"
#include "icode-common.h"
#include "codeproto.h"
#include "soft.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "gc.h"
#include "flags.h"
#include "code-analyse.h"
#include "machine.h"
#include "md.h"

extern int32 pc;
extern int32 npc;
extern int32 ppc;
extern int stackno;
extern int maxStack;
extern int maxLocal;
extern int maxArgs;
extern int maxTemp;
extern int isStatic;
extern int issuenr;

static int argcount;

#define	MAXLABTAB	64
label* labtab[MAXLABTAB];

#if defined(TWO_OPERAND)
#define	TWO_OPERAND_BEGIN1(_MOVE)		\
	SlotInfo* olddst = 0;			\
	if (src2 == dst) {			\
		olddst = dst;			\
		slot_alloctmp(dst);		\
	}					\
	if (src != dst) {			\
		_MOVE##(dst, src);		\
		src = dst;			\
	}
#define	TWO_OPERAND_BEGIN2(_MOVE)		\
	if (src != dst) {			\
		_MOVE##(dst, src);		\
		src = dst;			\
	}
#define	TWO_OPERAND_END1(_MOVE)			\
	if (olddst != 0) {			\
		_MOVE##(olddst, dst);		\
		slot_freetmp(dst);		\
	}
#else
#define	TWO_OPERAND_BEGIN1(_MOVE)
#define	TWO_OPERAND_BEGIN2(_MOVE)
#define	TWO_OPERAND_END1(_MOVE)
#endif

#define	NULLSLOT	0


/* ----------------------------------------------------------------------- */
/* Register loads and spills.						   */
/*									   */

#if defined(HAVE_spill_int)
void
spill_int(SlotInfo* src)
{
        _CALL_GEN_FUNC(spill_int)(src, SLOT2FRAMEOFFSET(src));
}
#endif

#if defined(HAVE_reload_int)
void
reload_int(SlotInfo* dst)
{
        _CALL_GEN_FUNC(reload_int)(dst, SLOT2FRAMEOFFSET(dst));
}
#endif

#if defined(HAVE_spill_float)
void
spill_float(SlotInfo* src)
{
        _CALL_GEN_FUNC(spill_float)(src, SLOT2FRAMEOFFSET(src));
}
#endif

#if defined(HAVE_reload_float)
void
reload_float(SlotInfo* dst)
{
        _CALL_GEN_FUNC(reload_float)(dst, SLOT2FRAMEOFFSET(dst));
}
#endif

#if defined(HAVE_spill_double)
void
spill_double(SlotInfo* src)
{
        _CALL_GEN_FUNC(spill_double)(src, SLOT2FRAMEOFFSET(src));
}
#endif

#if defined(HAVE_reload_double)
void
reload_double(SlotInfo* dst)
{
        _CALL_GEN_FUNC(reload_double)(dst, SLOT2FRAMEOFFSET(dst));
}
#endif

/*
 * Spill a slot in the correct way.
 */
void
spill(SlotInfo* src)
{
	if (src == 0 || src->reg == 0) {
		return;
	}
	if ((src->reg->modified & rwrite) != rwrite) {
		src->reg->modified = 0;
		return;
	}

	switch (src->reg->ctype) {
#if defined(HAVE_spill_int)
	case Rint:
#if defined(HAVE_ref_eq_int)
	case Rref:
	case Raddr:
#endif
		spill_int(src);
		break;
#endif
#if defined(HAVE_spill_long)
	case Rlong:
#if defined(HAVE_ref_eq_long)
	case Rref:
	case Raddr:
#endif
		spill_long(src);
		break;
#endif
#if defined(HAVE_spill_float)
	case Rfloat:
		spill_float(src);
		break;
#endif
#if defined(HAVE_spill_double)
	case Rdouble:
		spill_double(src);
		break;
#endif
	case Rany:
	case Ranylong:
	case Rvoid:
	default:
		assert("Failed to spill register" == 0);
	}

	src->reg->modified = 0;
}

/*
 * Reload a slot in the correct way.
 */
void
reload(SlotInfo* src)
{
	assert(src != 0);
	assert(src->reg != 0);

	switch (src->reg->ctype) {
#if defined(HAVE_reload_int)
	case Rint:
#if defined(HAVE_ref_eq_int)
	case Rref:
	case Raddr:
#endif
		reload_int(src);
		break;
#endif
#if defined(HAVE_reload_long)
	case Rlong:
#if defined(HAVE_ref_eq_long)
	case Rref:
	case Raddr:
#endif
		reload_long(src);
		break;
#endif
#if defined(HAVE_reload_float)
	case Rfloat:
		reload_float(src);
		break;
#endif
#if defined(HAVE_reload_double)
	case Rdouble:
		reload_double(src);
		break;
#endif
	case Rany:
	case Ranylong:
	case Rvoid:
	default:
		assert("Failed to spill register" == 0);
	}

	src->reg->modified = 0;
}

/* ----------------------------------------------------------------------- */
/* Prologues and epilogues.						   */
/*									   */

void
prologue(void)
{
	/* Start a block */
	start_basic_block();

	/* Emit prologue code */
        CALL_GEN_FUNC(prologue)();
}

void
epilogue(void)
{
        CALL_GEN_FUNC(epilogue)();
}


/* ----------------------------------------------------------------------- */
/* Conditional monitor management.					   */
/*									   */

void
mon_enter(methods* meth, SlotInfo* obj)
{
	/* Emit monitor entry if required */
        if ((meth->accflags & ACC_SYNCHRONISED) != 0) {
		prepare_function_call();
		if ((meth->accflags & ACC_STATIC) != 0) {
			pusharg_ref_const(meth->class);
		}
		else {
			pusharg_ref(obj);
		}
		call_ref((uintp)soft_monitorenter);
		popargs();
		fixup_function_call();
		reload_inner();
	}
}

void
mon_exit(methods* meth, SlotInfo* obj)
{
	/* Emit monitor exit if required */
        if ((meth->accflags & ACC_SYNCHRONISED) != 0) {
		prepare_function_call();
		if ((meth->accflags & ACC_STATIC) != 0) {
			pusharg_ref_const(meth->class);
		}
		else {
			pusharg_ref(obj);
		}
		call_ref((uintp)soft_monitorexit);
		popargs();
		fixup_function_call();
		reload_inner();
	}
}

/* ----------------------------------------------------------------------- */
/* Basic block and instruction management.				   */
/*									   */

void
_start_exception_block(SlotInfo* topofstack)
{
	CALL_GEN_FUNC(exception_prologue)();
	return_ref(topofstack);
}


/* ----------------------------------------------------------------------- */
/* Moves.								   */
/*									   */

#if defined(HAVE_move_int_const)
void
move_int_const(SlotInfo* dst, jint val)
{
        CALL_GEN_FUNC(move_int_const)(dst, NULLSLOT, val);
}
#endif

void
move_long_const(SlotInfo* dst, jlong val)
{
#if defined(HAVE_move_long_const)
        CALL_GEN_FUNC(move_long_const)(dst, NULLSLOT, val);
	dst->v.tint = val;
#else
	move_int_const(dst, (jint)(val & 0xFFFFFFFF));
	move_int_const(dst+1, (jint)((val >> 32) & 0xFFFFFFFF));
#endif
}

#if defined(HAVE_move_float_const)
void
move_float_const(SlotInfo* dst, float val)
{
        CALL_GEN_FUNC(move_float_const)(dst, NULLSLOT, val);
}
#endif

#if defined(HAVE_move_double_const)
void
move_double_const(SlotInfo* dst, jdouble val)
{
        CALL_GEN_FUNC(move_double_const)(dst, NULLSLOT, val);
}
#endif

#if defined(HAVE_move_int)
void
move_int(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(move_int)(dst, NULLSLOT, src);
}
#endif

void
move_long(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_move_long)
        CALL_GEN_FUNC(move_long)(dst, NULLSLOT, src);
#else
	assert(dst != src+1);
	move_int(dst, src);
	move_int(dst+1, src+1);
#endif
}

#if defined(HAVE_move_float)
void
move_float(SlotInfo* dst, SlotInfo* src)
{
        CALL_GEN_FUNC(move_float)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_move_double)
void
move_double(SlotInfo* dst, SlotInfo* src)
{
        CALL_GEN_FUNC(move_double)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_move_label_const)
void
move_label_const(SlotInfo* dst, label* lab)
{
        CALL_GEN_FUNC(move_label_const)(dst, NULLSLOT, (int)lab);
}
#endif

void
swap_int(SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_swap_int)
        CALL_GEN_FUNC(swap_int)(src, NULLSLOT, src2);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int(tmp, src);
	move_int(src, src2);
	move_int(src2, tmp);
	slot_freetmp(tmp);
#endif
}

void
swap_long(SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_swap_long)
        CALL_GEN_FUNC(swap_long)(src, NULLSLOT, src2);
#else
	swap_int(src, src2);
	swap_int(src+1, src2+1);
#endif
}



/* ----------------------------------------------------------------------- */
/* Arithmetic operators - add, sub, etc.				   */
/*									   */

void
adc_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_adc_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(adc_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	adc_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_adc_int)
void
adc_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_adc_int_const)
	if (rslot_isconst(src2)) {
		adc_int_const(dst, src, rslot_value(src2));
		return;
	}
	if (rslot_isconst(src)) {
		adc_int_const(dst, src2, rslot_value(src));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(adc_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
add_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_add_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(add_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	add_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_add_int)
void
add_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_add_int_const)
	if (rslot_isconst(src2)) {
		add_int_const(dst, src, rslot_value(src2));
		return;
	}
	if (rslot_isconst(src)) {
		add_int_const(dst, src2, rslot_value(src));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(add_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
add_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_add_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(add_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	add_int(dst, src, src2);
	adc_int(dst+1, src+1, src2+1);
#endif
}

#if defined(HAVE_add_float)
void
add_float(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_float);
        CALL_GEN_FUNC(add_float)(dst, src, src2);
	TWO_OPERAND_END1(move_float);
}
#endif

#if defined(HAVE_add_double)
void
add_double(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_double);
        CALL_GEN_FUNC(add_double)(dst, src, src2);
	TWO_OPERAND_END1(move_double);
}
#endif

#if defined(HAVE_sbc_int)
void
sbc_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_int);
        CALL_GEN_FUNC(sbc_int)(dst, src, src2);
	TWO_OPERAND_END1(move_int);
}
#endif

void
sub_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_sub_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(sub_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	sub_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_sub_int)
void
sub_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_sub_int_const)
	if (rslot_isconst(src2)) {
		sub_int_const(dst, src, rslot_value(src2));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(sub_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
sub_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_sub_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(sub_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	sub_int(dst, src, src2);
	sbc_int(dst+1, src+1, src2+1);
#endif
}

#if defined(HAVE_sub_float)
void
sub_float(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_float);
        CALL_GEN_FUNC(sub_float)(dst, src, src2);
	TWO_OPERAND_END1(move_float);
}
#endif

#if defined(HAVE_sub_double)
void
sub_double(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_double);
        CALL_GEN_FUNC(sub_double)(dst, src, src2);
	TWO_OPERAND_END1(move_double);
}
#endif

void
mul_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_mul_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(mul_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	mul_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_mul_int)
void
mul_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_mul_int_const)
	if (rslot_isconst(src2)) {
		mul_int_const(dst, src, rslot_value(src2));
		return;
	}
	if (rslot_isconst(src)) {
		mul_int_const(dst, src2, rslot_value(src));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(mul_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
mul_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_mul_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(mul_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	prepare_function_call();
	pusharg_long(src2);
	pusharg_long(src);
	call_ref((uintp)soft_lmul);
	popargs();
	fixup_function_call();
	return_long(dst);
	reload_inner();
#endif
}

#if defined(HAVE_mul_float)
void
mul_float(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_float);
        CALL_GEN_FUNC(mul_float)(dst, src, src2);
	TWO_OPERAND_END1(move_float);
}
#endif

#if defined(HAVE_mul_double)
void
mul_double(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_double);
        CALL_GEN_FUNC(mul_double)(dst, src, src2);
	TWO_OPERAND_END1(move_double);
}
#endif

void
div_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_div_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(div_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	div_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_div_int)
void
div_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_int);
	CALL_GEN_FUNC(div_int)(dst, src, src2);
	TWO_OPERAND_END1(move_int);
}
#endif

void
div_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_div_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(div_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	prepare_function_call();
	pusharg_long(src2);
	pusharg_long(src);
	call_ref((uintp)soft_ldiv);
	popargs();
	fixup_function_call();
	return_long(dst);
	reload_inner();
#endif
}

void
div_float(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_div_float)
	TWO_OPERAND_BEGIN1(move_float);
        CALL_GEN_FUNC(div_float)(dst, src, src2);
	TWO_OPERAND_END1(move_float);
#else
	prepare_function_call();
	pusharg_float(src2);
	pusharg_float(src);
	call_ref((uintp)soft_fdiv);
	popargs();
	fixup_function_call();
	return_float(dst);
	reload_inner();
#endif
}

void
div_double(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_div_double)
	TWO_OPERAND_BEGIN1(move_double);
        CALL_GEN_FUNC(div_double)(dst, src, src2);
	TWO_OPERAND_END1(move_double);
#else
	prepare_function_call();
	pusharg_double(src2);
	pusharg_double(src);
	call_ref((uintp)soft_fdivl);
	popargs();
	fixup_function_call();
	return_double(dst);
	reload_inner();
#endif
}

void
rem_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_rem_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(rem_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	rem_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_rem_int)
void
rem_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
	TWO_OPERAND_BEGIN1(move_int);
	CALL_GEN_FUNC(rem_int)(dst, src, src2);
	TWO_OPERAND_END1(move_int);
}
#endif

void
rem_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_rem_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(rem_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	prepare_function_call();
	pusharg_long(src2);
	pusharg_long(src);
	call_ref((uintp)soft_lrem);
	popargs();
	fixup_function_call();
	return_long(dst);
	reload_inner();
#endif
}

void
rem_float(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_rem_float)
	TWO_OPERAND_BEGIN1(move_float);
        CALL_GEN_FUNC(rem_float)(dst, src, src2);
	TWO_OPERAND_END1(move_float);
#else
	prepare_function_call();
	pusharg_float(src2);
	pusharg_float(src);
	call_ref((uintp)soft_frem);
	popargs();
	fixup_function_call();
	return_float(dst);
	reload_inner();
#endif
}

void
rem_double(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_rem_double)
	TWO_OPERAND_BEGIN1(move_double);
        CALL_GEN_FUNC(rem_double)(dst, src, src2);
	TWO_OPERAND_END1(move_double);
#else
	prepare_function_call();
	pusharg_double(src2);
	pusharg_double(src);
	call_ref((uintp)soft_freml);
	popargs();
	fixup_function_call();
	return_double(dst);
	reload_inner();
#endif
}

void
neg_int(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_neg_int)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(neg_int)(dst, NULLSLOT, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, 0);
	sub_int(dst, tmp, src);
	slot_freetmp(tmp);
#endif
}

void
neg_long(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_neg_long)
	TWO_OPERAND_BEGIN2(move_long);
        CALL_GEN_FUNC(neg_long)(dst, NULLSLOT, src);
#elif defined(HAVE_sbc_int)
	SlotInfo* zero;
	slot_alloctmp(zero);
	move_int_const(zero, 0);
	sub_int(dst, zero, src);
	sbc_int(dst+1, zero, src+1);
	slot_freetmp(zero);
#elif defined(HAVE_adc_int_const)
	neg_int(dst, src);
	adc_int_const(dst+1, src+1, 0);
	neg_int(dst+1, dst+1);
#else
	abort();
#endif
}

void
neg_float(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_neg_float)
	TWO_OPERAND_BEGIN2(move_float);
        CALL_GEN_FUNC(neg_float)(dst, NULLSLOT, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_float_const(tmp, 0);
	sub_float(dst, tmp, src);
	slot_freetmp(tmp);
#endif
}

void
neg_double(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_neg_double)
	TWO_OPERAND_BEGIN2(move_double);
        CALL_GEN_FUNC(neg_double)(dst, NULLSLOT, src);
#else
	SlotInfo* tmp;
	slot_alloc2tmp(tmp);
	move_double_const(tmp, 0);
	sub_double(dst, tmp, src);
	slot_free2tmp(tmp);
#endif
}


/* ----------------------------------------------------------------------- */
/* Logical operators - and, or, etc.					   */
/*									   */

void
and_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_and_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(and_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	and_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_and_int)
void
and_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_and_int_const)
	if (rslot_isconst(src2)) {
		and_int_const(dst, src, rslot_value(src2));
		return;
	}
	if (rslot_isconst(src)) {
		and_int_const(dst, src2, rslot_value(src));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(and_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
and_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_and_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(and_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	and_int(dst, src, src2);
	and_int(dst+1, src+1, src2+1);
#endif
}

void
or_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_or_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(or_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	or_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_or_int)
void
or_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_or_int_const)
	if (rslot_isconst(src2)) {
		or_int_const(dst, src, rslot_value(src2));
		return;
	}
	if (rslot_isconst(src)) {
		or_int_const(dst, src2, rslot_value(src));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(or_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
or_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_or_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(or_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	or_int(dst, src, src2);
	or_int(dst+1, src+1, src2+1);
#endif
}

void
xor_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_xor_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(xor_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	xor_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_xor_int)
void
xor_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_xor_int_const)
	if (rslot_isconst(src2)) {
		xor_int_const(dst, src, rslot_value(src2));
		return;
	}
	if (rslot_isconst(src)) {
		xor_int_const(dst, src2, rslot_value(src));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(xor_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
xor_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_xor_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(xor_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	xor_int(dst, src, src2);
	xor_int(dst+1, src+1, src2+1);
#endif
}

void
lshl_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_lshl_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(lshl_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	lshl_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_lshl_int)
void
lshl_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_lshl_int_const)
	if (rslot_isconst(src2)) {
		lshl_int_const(dst, src, rslot_value(src2));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(lshl_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
lshl_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_lshl_long)
	TWO_OPERAND_BEGIN1(move_long);
        CALL_GEN_FUNC(lshl_long)(dst, src, src2);
	TWO_OPERAND_END1(move_long);
#else
	prepare_function_call();
	pusharg_int(src2);
	pusharg_long(src);
	call_ref((uintp)soft_lshll);
	popargs();
	fixup_function_call();
	return_long(dst);
	reload_inner();
#endif
}

void
ashr_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_ashr_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(ashr_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	ashr_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_ashr_int)
void
ashr_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_ashr_int_const)
	if (rslot_isconst(src2)) {
		ashr_int_const(dst, src, rslot_value(src2));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(ashr_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
ashr_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_ashr_long)
	TWO_OPERAND_BEGIN2(move_long);
        CALL_GEN_FUNC(ashr_long)(dst, src, src2);
#else
	prepare_function_call();
	pusharg_int(src2);
	pusharg_long(src);
	call_ref((uintp)soft_ashrl);
	popargs();
	fixup_function_call();
	return_long(dst);
	reload_inner();
#endif
}

void
lshr_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_lshr_int_const)
	TWO_OPERAND_BEGIN2(move_int);
        CALL_GEN_FUNC(lshr_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	lshr_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_lshr_int)
void
lshr_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_lshr_int_const)
	if (rslot_isconst(src2)) {
		lshr_int_const(dst, src, rslot_value(src2));
		return;
	}
#endif
	{
		TWO_OPERAND_BEGIN1(move_int);
		CALL_GEN_FUNC(lshr_int)(dst, src, src2);
		TWO_OPERAND_END1(move_int);
	}
}
#endif

void
lshr_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_lshr_long)
	TWO_OPERAND_BEGIN2(move_long);
        CALL_GEN_FUNC(lshr_long)(dst, src, src2);
#else
	prepare_function_call();
	pusharg_int(src2);
	pusharg_long(src);
	call_ref((uintp)soft_lshrl);
	popargs();
	fixup_function_call();
	return_long(dst);
	reload_inner();
#endif
}




/* ----------------------------------------------------------------------- */
/* Load and store.							   */
/*									   */

#if defined(HAVE_load_int)
void
load_int(SlotInfo* dst, SlotInfo* src)
{
        CALL_GEN_FUNC(load_int)(dst, NULLSLOT, src);
}
#endif

void
load_long(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_load_long)
	CALL_GEN_FUNC(load_long)(dst, NULLSLOT, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	add_ref_const(tmp, src, 4);
	load_int(dst, src);
	load_int(dst+1, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_load_float)
void
load_float(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(load_float)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_load_double)
void
load_double(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(load_double)(dst, NULLSLOT, src);
}
#endif

void
load_byte(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_load_byte)
	CALL_GEN_FUNC(load_byte)(dst, NULLSLOT, src);
#else
	load_int(dst, src);
	lshl_int_const(dst, dst, 8 * (sizeof(jint) - sizeof(jbyte)));
	ashr_int_const(dst, dst, 8 * (sizeof(jint) - sizeof(jbyte)));
#endif
}

void
load_char(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_load_char)
	CALL_GEN_FUNC(load_char)(dst, NULLSLOT, src);
#else
	load_int(dst, src);
	and_int_const(dst, dst, (1 << (8 * sizeof(jchar))) - 1);
#endif
}

void
load_short(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_load_short)
	CALL_GEN_FUNC(load_short)(dst, NULLSLOT, src);
#else
	load_int(dst, src);
	lshl_int_const(dst, dst, 8 * (sizeof(jint) - sizeof(jshort)));
	ashr_int_const(dst, dst, 8 * (sizeof(jint) - sizeof(jshort)));
#endif
}

void
load_code_ref(SlotInfo* dst, SlotInfo* src)
{
	load_ref(dst, src);
}

void
load_key(SlotInfo* dst, SlotInfo* src)
{
	load_int(dst, src);
}

void
load_addr_int(SlotInfo* dst, uintp src)
{
#if defined(HAVE_load_addr_int)
        CALL_GEN_FUNC(load_addr_int)(dst, NULLSLOT, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_ref_const(tmp, src);
	load_int(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_addr_long(SlotInfo* dst, uintp src)
{
#if defined(HAVE_load_addr_long)
        CALL_GEN_FUNC(load_addr_long)(dst, NULLSLOT, src);
#else
	load_addr_int(dst, src);
	load_addr_int(dst+1, src+4);
#endif
}

void
load_addr_float(SlotInfo* dst, uintp src)
{
#if defined(HAVE_load_addr_float)
        CALL_GEN_FUNC(load_addr_float)(dst, NULLSLOT, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_ref_const(tmp, src);
	load_float(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_addr_double(SlotInfo* dst, uintp src)
{
#if defined(HAVE_load_addr_double)
        CALL_GEN_FUNC(load_addr_double)(dst, NULLSLOT, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_ref_const(tmp, src);
	load_double(dst, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_store_int)
void
store_int(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(store_int)(NULLSLOT, dst, src);
}
#endif

void
store_long(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_store_long)
	CALL_GEN_FUNC(store_long)(NULLSLOT, dst, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	add_ref_const(tmp, dst, 4);
	store_int(dst, src);
	store_int(tmp, src+1);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_store_float)
void
store_float(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(store_float)(NULLSLOT, dst, src);
}
#endif

#if defined(HAVE_store_double)
void
store_double(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(store_double)(NULLSLOT, dst, src);
}
#endif

void
store_byte(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_store_byte)
	CALL_GEN_FUNC(store_byte)(NULLSLOT, dst, src);
#else
	SlotInfo* tmp;
	SlotInfo* tmp2;
	slot_alloctmp(tmp);
	slot_alloctmp(tmp2);
	and_int_const(tmp, src, (1 << (8 * sizeof(jbyte))) - 1);
	load_int(tmp2, dst);
	and_int_const(tmp2, tmp2, -(1 << (8 * sizeof(jbyte))));
	or_int(tmp2, tmp2, tmp);
	store_int(dst, tmp2);
	slot_freetmp(tmp2);
	slot_freetmp(tmp);
#endif
}

void
store_char(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_store_char)
	CALL_GEN_FUNC(store_char)(NULLSLOT, dst, src);
#else
	SlotInfo* tmp;
	SlotInfo* tmp2;
	slot_alloctmp(tmp);
	slot_alloctmp(tmp2);
	and_int_const(tmp, src, (1 << (8 * sizeof(jchar))) - 1);
	load_int(tmp2, dst);
	and_int_const(tmp2, tmp2, -(1 << (8 * sizeof(jchar))));
	or_int(tmp2, tmp2, tmp);
	store_int(dst, tmp2);
	slot_freetmp(tmp2);
	slot_freetmp(tmp);
#endif
}

void
store_short(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_store_short)
	CALL_GEN_FUNC(store_short)(NULLSLOT, dst, src);
#else
	SlotInfo* tmp;
	SlotInfo* tmp2;
	slot_alloctmp(tmp);
	slot_alloctmp(tmp2);
	and_int_const(tmp, src, (1 << (8 * sizeof(jshort))) - 1);
	load_int(tmp2, dst);
	and_int_const(tmp2, tmp2, -(1 << (8 * sizeof(jshort))));
	or_int(tmp2, tmp2, tmp);
	store_int(dst, tmp2);
	slot_freetmp(tmp2);
	slot_freetmp(tmp);
#endif
}

void
store_addr_int(uintp dst, SlotInfo* src)
{
#if defined(HAVE_store_addr_int)
	CALL_GEN_FUNC(store_addr_int)(NULLSLOT, dst, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_ref_const(tmp, dst);
	store_int(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_addr_long(uintp dst, SlotInfo* src)
{
#if defined(HAVE_store_addr_long)
	CALL_GEN_FUNC(store_addr_long)(NULLSLOT, dst, src);
#else
	store_addr_int(dst, src);
	store_addr_int(dst+4, src+1);
#endif
}

void
store_addr_float(uintp dst, SlotInfo* src)
{
#if defined(HAVE_store_addr_float)
	CALL_GEN_FUNC(store_addr_float)(NULLSLOT, dst, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_ref_const(tmp, dst);
	store_float(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_addr_double(uintp dst, SlotInfo* src)
{
#if defined(HAVE_store_addr_double)
	CALL_GEN_FUNC(store_addr_double)(NULLSLOT, dst, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_ref_const(tmp, dst);
	store_double(tmp, src);
	slot_freetmp(tmp);
#endif
}

/* ----------------------------------------------------------------------- */
/* Load and store extensions						   */
/*									   */

void
load_offset_int(SlotInfo* dst, SlotInfo* src, int offset)
{
	if (offset == 0) {
		load_int(dst, src);
		return;
	}
#if defined(HAVE_load_offset_int)
	CALL_GEN_FUNC(load_offset_int)(dst, src, offset);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	add_ref_const(tmp, src, offset);
	load_int(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_offset_long(SlotInfo* dst, SlotInfo* src, int offset)
{
	SlotInfo* tmp;
	if (offset == 0) {
		load_long(dst, src);
		return;
	}
#if defined(HAVE_load_offset_long)
	CALL_GEN_FUNC(load_offset_long)(dst, src, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, src, offset);
	load_long(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_offset_float(SlotInfo* dst, SlotInfo* src, int offset)
{
	SlotInfo* tmp;
	if (offset == 0) {
		load_float(dst, src);
		return;
	}
#if defined(HAVE_load_offset_float)
	CALL_GEN_FUNC(load_offset_float)(dst, src, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, src, offset);
	load_float(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_offset_double(SlotInfo* dst, SlotInfo* src, int offset)
{
	SlotInfo* tmp;
	if (offset == 0) {
		load_double(dst, src);
		return;
	}
#if defined(HAVE_load_offset_double)
	CALL_GEN_FUNC(load_offset_double)(dst, src, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, src, offset);
	load_double(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_offset_byte(SlotInfo* dst, SlotInfo* src, int offset)
{
	SlotInfo* tmp;
	if (offset == 0) {
		load_byte(dst, src);
		return;
	}
#if defined(HAVE_load_offset_byte)
	CALL_GEN_FUNC(load_offset_byte)(dst, src, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, src, offset);
	load_byte(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_offset_char(SlotInfo* dst, SlotInfo* src, int offset)
{
	SlotInfo* tmp;
	if (offset == 0) {
		load_char(dst, src);
		return;
	}
#if defined(HAVE_load_offset_char)
	CALL_GEN_FUNC(load_offset_char)(dst, src, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, src, offset);
	load_char(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_offset_short(SlotInfo* dst, SlotInfo* src, int offset)
{
	SlotInfo* tmp;
	if (offset == 0) {
		load_short(dst, src);
		return;
	}
#if defined(HAVE_load_offset_short)
	CALL_GEN_FUNC(load_offset_short)(dst, src, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, src, offset);
	load_short(dst, tmp);
	slot_freetmp(tmp);
#endif
}

void
load_offset_scaled_int(SlotInfo* dst, SlotInfo* src, SlotInfo* idx, int offset)
{
	SlotInfo* tmp;
#if defined(HAVE_load_offset_scaled_int)
	CALL_GEN_FUNC(load_offset_scaled_int)(dst, src, idx, offset);
#else
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jint);
	add_ref(tmp, tmp, src);
	load_offset_int(dst, tmp, offset);
	slot_freetmp(tmp);
#endif
}

void
load_offset_scaled_long(SlotInfo* dst, SlotInfo* src, SlotInfo* idx, int offset)
{
	SlotInfo* tmp;
#if defined(HAVE_load_offset_scaled_long)
	CALL_GEN_FUNC(load_offset_scaled_long)(dst, src, idx, offset);
#else
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jlong);
	add_ref(tmp, tmp, src);
	load_offset_long(dst, tmp, offset);
	slot_freetmp(tmp);
#endif
}

void
load_offset_scaled_float(SlotInfo* dst, SlotInfo* src, SlotInfo* idx, int offset)
{
	SlotInfo* tmp;
#if defined(HAVE_load_offset_scaled_float)
	CALL_GEN_FUNC(load_offset_scaled_float)(dst, src, idx, offset);
#else
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jfloat);
	add_ref(tmp, tmp, src);
	load_offset_float(dst, tmp, offset);
	slot_freetmp(tmp);
#endif
}

void
load_offset_scaled_double(SlotInfo* dst, SlotInfo* src, SlotInfo* idx, int offset)
{
	SlotInfo* tmp;
#if defined(HAVE_load_offset_scaled_double)
	CALL_GEN_FUNC(load_offset_scaled_double)(dst, src, idx, offset);
#else
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jdouble);
	add_ref(tmp, tmp, src);
	load_offset_double(dst, tmp, offset);
	slot_freetmp(tmp);
#endif
}

void
load_offset_scaled_byte(SlotInfo* dst, SlotInfo* src, SlotInfo* idx, int offset)
{
	SlotInfo* tmp;
#if defined(HAVE_load_offset_scaled_byte)
	CALL_GEN_FUNC(load_offset_scaled_byte)(dst, src, idx, offset);
#else
	slot_alloctmp(tmp);
	if (SHIFT_jbyte > 0) {
		lshl_int_const(tmp, idx, SHIFT_jbyte);
		add_ref(tmp, tmp, src);
	}
	else {
		add_ref(tmp, idx, src);
	}
	load_offset_byte(dst, tmp, offset);
	slot_freetmp(tmp);
#endif
}

void
load_offset_scaled_char(SlotInfo* dst, SlotInfo* src, SlotInfo* idx, int offset)
{
	SlotInfo* tmp;
#if defined(HAVE_load_offset_scaled_char)
	CALL_GEN_FUNC(load_offset_scaled_char)(dst, src, idx, offset);
#else
	slot_alloctmp(tmp);
	if (SHIFT_jchar > 0) {
		lshl_int_const(tmp, idx, SHIFT_jchar);
		add_ref(tmp, tmp, src);
	}
	else {
		add_ref(tmp, idx, src);
	}
	load_offset_char(dst, tmp, offset);
	slot_freetmp(tmp);
#endif
}

void
load_offset_scaled_short(SlotInfo* dst, SlotInfo* src, SlotInfo* idx, int offset)
{
	SlotInfo* tmp;
#if defined(HAVE_load_offset_scaled_short)
	CALL_GEN_FUNC(load_offset_scaled_short)(dst, src, idx, offset);
#else
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jshort);
	add_ref(tmp, tmp, src);
	load_offset_short(dst, tmp, offset);
	slot_freetmp(tmp);
#endif
}

void
store_offset_int(SlotInfo* dst, int offset, SlotInfo* src)
{
	SlotInfo* tmp;
	if (offset == 0) {
		store_int(dst, src);
		return;
	}
#if defined(HAVE_store_offset_int)
	CALL_GEN_FUNC(store_offset_int)(src, dst, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, dst, offset);
	store_int(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_long(SlotInfo* dst, int offset, SlotInfo* src)
{
	SlotInfo* tmp;
	if (offset == 0) {
		store_long(dst, src);
		return;
	}
#if defined(HAVE_store_offset_long)
	CALL_GEN_FUNC(store_offset_long)(src, dst, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, dst, offset);
	store_long(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_float(SlotInfo* dst, int offset, SlotInfo* src)
{
	SlotInfo* tmp;
	if (offset == 0) {
		store_float(dst, src);
		return;
	}
#if defined(HAVE_store_offset_float)
	CALL_GEN_FUNC(store_offset_float)(src, dst, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, dst, offset);
	store_float(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_double(SlotInfo* dst, int offset, SlotInfo* src)
{
	SlotInfo* tmp;
	if (offset == 0) {
		store_double(dst, src);
		return;
	}
#if defined(HAVE_store_offset_double)
	CALL_GEN_FUNC(store_offset_double)(src, dst, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, dst, offset);
	store_double(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_byte(SlotInfo* dst, int offset, SlotInfo* src)
{
	SlotInfo* tmp;
	if (offset == 0) {
		store_byte(dst, src);
		return;
	}
#if defined(HAVE_store_offset_byte)
	CALL_GEN_FUNC(store_offset_byte)(src, dst, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, dst, offset);
	store_byte(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_char(SlotInfo* dst, int offset, SlotInfo* src)
{
	SlotInfo* tmp;
	if (offset == 0) {
		store_char(dst, src);
		return;
	}
#if defined(HAVE_store_offset_char)
	CALL_GEN_FUNC(store_offset_char)(src, dst, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, dst, offset);
	store_char(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_short(SlotInfo* dst, int offset, SlotInfo* src)
{
	SlotInfo* tmp;
	if (offset == 0) {
		store_short(dst, src);
		return;
	}
#if defined(HAVE_store_offset_short)
	CALL_GEN_FUNC(store_offset_short)(src, dst, offset);
#else
	slot_alloctmp(tmp);
	add_ref_const(tmp, dst, offset);
	store_short(tmp, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_scaled_int(SlotInfo* dst, SlotInfo* idx, int offset, SlotInfo* src)
{
#if defined(HAVE_store_offset_scaled_int)
	CALL_GEN_FUNC(store_offset_scaled_int)(dst, idx, offset, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jint);
	add_ref(tmp, tmp, dst);
	store_offset_int(tmp, offset, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_scaled_long(SlotInfo* dst, SlotInfo* idx, int offset, SlotInfo* src)
{
#if defined(HAVE_store_offset_scaled_long)
	CALL_GEN_FUNC(store_offset_scaled_long)(dst, idx, offset, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jlong);
	add_ref(tmp, tmp, dst);
	store_offset_long(tmp, offset, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_scaled_float(SlotInfo* dst, SlotInfo* idx, int offset, SlotInfo* src)
{
#if defined(HAVE_store_offset_scaled_float)
	CALL_GEN_FUNC(store_offset_scaled_float)(dst, idx, offset, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jfloat);
	add_ref(tmp, tmp, dst);
	store_offset_float(tmp, offset, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_scaled_double(SlotInfo* dst, SlotInfo* idx, int offset, SlotInfo* src)
{
#if defined(HAVE_store_offset_scaled_double)
	CALL_GEN_FUNC(store_offset_scaled_double)(dst, idx, offset, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jdouble);
	add_ref(tmp, tmp, dst);
	store_offset_double(tmp, offset, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_scaled_byte(SlotInfo* dst, SlotInfo* idx, int offset, SlotInfo* src)
{
#if defined(HAVE_store_offset_scaled_byte_const)
	if (rslot_isconst(src)) {
		CALL_GEN_FUNC(store_offset_scaled_byte_const)(dst, idx, offset, rslot_value(src));
		return;
	}
#endif
#if defined(HAVE_store_offset_scaled_byte)
	CALL_GEN_FUNC(store_offset_scaled_byte)(dst, idx, offset, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	if (SHIFT_jbyte > 0) {
		lshl_int_const(tmp, idx, SHIFT_jbyte);
		add_ref(tmp, tmp, dst);
	}
	else {
		add_ref(tmp, idx, dst);
	}
	store_offset_byte(tmp, offset, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_scaled_char(SlotInfo* dst, SlotInfo* idx, int offset, SlotInfo* src)
{
#if defined(HAVE_store_offset_scaled_char)
	CALL_GEN_FUNC(store_offset_scaled_char)(dst, idx, offset, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	if (SHIFT_jchar > 0) {
		lshl_int_const(tmp, idx, SHIFT_jchar);
		add_ref(tmp, tmp, dst);
	}
	else {
		add_ref(tmp, idx, dst);
	}
	store_offset_char(tmp, offset, src);
	slot_freetmp(tmp);
#endif
}

void
store_offset_scaled_short(SlotInfo* dst, SlotInfo* idx, int offset, SlotInfo* src)
{
#if defined(HAVE_store_offset_scaled_short)
	CALL_GEN_FUNC(store_offset_scaled_short)(dst, idx, offset, src);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	lshl_int_const(tmp, idx, SHIFT_jshort);
	add_ref(tmp, tmp, dst);
	store_offset_short(tmp, offset, src);
	slot_freetmp(tmp);
#endif
}


/* ----------------------------------------------------------------------- */
/* Function argument management.					   */
/*									   */

void
pusharg_int_const(int val)
{
#if defined(HAVE_pusharg_int_const)
	CALL_GEN_FUNC(pusharg_int_const)(NULLSLOT, NULLSLOT, val);
	argcount++;
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	pusharg_int(tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_pusharg_int_slot)
void
pusharg_int_slot(SlotInfo* src)
{
	CALL_GEN_FUNC(pusharg_int_slot)(NULLSLOT, NULLSLOT, src);
	argcount += 1;
}
#endif

#if defined(HAVE_pusharg_int)
void
pusharg_int(SlotInfo* src)
{
#if defined(HAVE_pusharg_int_const)
	if (rslot_isconst(src)) {
		pusharg_int_const(rslot_value(src));
		return;
	}
#endif
#if defined(HAVE_pusharg_int_slot)
	if (!register_is_allocated(src)) {
		pusharg_int_slot(src);
		return;
	}
#endif
	CALL_GEN_FUNC(pusharg_int)(NULLSLOT, NULLSLOT, src);
	argcount += 1;
}
#endif

#if defined(HAVE_pusharg_float)
void
pusharg_float(SlotInfo* src)
{
	CALL_GEN_FUNC(pusharg_float)(NULLSLOT, NULLSLOT, src);
	argcount += 1;
}
#endif

#if defined(HAVE_pusharg_double)
void
pusharg_double(SlotInfo* src)
{
	CALL_GEN_FUNC(pusharg_double)(NULLSLOT, NULLSLOT, src);
	argcount += 2;
}
#endif

void
pusharg_long(SlotInfo* src)
{
#if defined(HAVE_pusharg_long)
	CALL_GEN_FUNC(pusharg_long)(NULLSLOT, NULLSLOT, src);
	argcount += 2;
#else
	pusharg_int(src+1);
	pusharg_int(src);
#endif
}

#if defined(HAVE_popargs)
void
popargs(void)
{
	CALL_GEN_FUNC(popargs)(NULLSLOT, NULLSLOT, argcount);
	argcount = 0;
}
#endif



/* ----------------------------------------------------------------------- */
/* Control flow changes.						   */
/*									   */

#if defined(HAVE_branch)

void
branch(label* dst, int type)
{
	CALL_GEN_FUNC(branch)(NULLSLOT, (int)dst, type);
}
#endif

void
cbranch_int(SlotInfo* s1, SlotInfo* s2, label* dst, int type)
{
#if defined(HAVE_cbranch_int)
	CALL_GEN_FUNC(cbranch_int)(NULLSLOT, s1, s2, (int)dst, type);
#else
	cmp_int(0, s1, s2);
	branch(dst, type);
#endif
}

void
cbranch_int_const(SlotInfo* s1, jint s2, label* dst, int type)
{
#if defined(HAVE_cbranch_int_const)
	CALL_GEN_FUNC(cbranch_int_const)(NULLSLOT, s1, s2, (int)dst, type);
#else
	cmp_int_const(0, s1, s2);
	branch(dst, type);
#endif
}

cbranch_load_offset_int(SlotInfo* s1, SlotInfo* s2, int o, label* dst, int type)
{
#if defined(HAVE_cbranch_load_offset_int)
	CALL_GEN_FUNC(cbranch_load_offset_int)(NULLSLOT, s1, s2, o, (int)dst, type);
#else
	cmp_load_offset_int(0, s1, s2, o);
	branch(dst, type);
#endif
}

#if defined(HAVE_branch_indirect)
void
branch_indirect(SlotInfo* dst)
{
	CALL_GEN_FUNC(branch_indirect)(NULLSLOT, dst, ba);
}
#endif

#if defined(HAVE_call_ref)
void
call_ref(uintp routine)
{
	label* l = nextLabel();
	l->to = routine;	/* What place does it goto */
	l->type = Labsolute|Lexternal;
	l->at = 0;
	l->from = 0;
	CALL_GEN_FUNC(call_ref)(NULLSLOT, (int)l, ba);
}
#endif

#if defined(HAVE_call)
void
call(SlotInfo* dst, SlotInfo* ret)
{
	slot_nowriteback(dst);
	CALL_GEN_FUNC(call)(NULLSLOT, dst, ba);
}
#endif

#if defined(HAVE_ret)
void
ret(void)
{
	CALL_GEN_FUNC(ret)(NULLSLOT, NULLSLOT, NULLSLOT);
}
#endif

#if defined(HAVE_returnarg_int)
void
returnarg_int(SlotInfo* src)
{
	CALL_GEN_FUNC(returnarg_int)(&returninfo[returnInt], NULLSLOT, src);
}
#endif

#if defined(HAVE_returnarg_long)
void
returnarg_long(SlotInfo* src)
{
	CALL_GEN_FUNC(returnarg_long)(&returninfo[returnLong], NULLSLOT, src);
}
#endif

#if defined(HAVE_returnarg_float)
void
returnarg_float(SlotInfo* src)
{
	CALL_GEN_FUNC(returnarg_float)(&returninfo[returnFloat], NULLSLOT, src);
}
#endif

#if defined(HAVE_returnarg_double)
void
returnarg_double(SlotInfo* src)
{
	CALL_GEN_FUNC(returnarg_double)(&returninfo[returnDouble], NULLSLOT, src);
}
#endif

#if defined(HAVE_return_int)
void
return_int(SlotInfo* dst)
{
	CALL_GEN_FUNC(return_int)(dst, NULLSLOT, NULLSLOT);

}
#endif

#if defined(HAVE_return_long)
void
return_long(SlotInfo* dst)
{
	CALL_GEN_FUNC(return_long)(dst, NULLSLOT, NULLSLOT);
}
#endif

#if defined(HAVE_return_float)
void
return_float(SlotInfo* dst)
{
	CALL_GEN_FUNC(return_float)(dst, NULLSLOT, NULLSLOT);
}
#endif

#if defined(HAVE_return_double)
void
return_double(SlotInfo* dst)
{
	CALL_GEN_FUNC(return_double)(dst, NULLSLOT, NULLSLOT);
}
#endif



/* ----------------------------------------------------------------------- */
/* Labels.								   */
/*									   */

label*
reference_label(int32 i, int32 n)
{
	label* l;

	assert(n < MAXLABTAB);
	if (labtab[n] == 0) {
		l = nextLabel();
		labtab[n] = l;
		l->type = Lnull;
		l->at = 0;
		l->from = 0;
		l->to = 0;
	}
	else {
		l = labtab[n];
		labtab[n] = 0;
	}
	return (l);
}

label*
reference_code_label(int32 offset)
{
	label* l = nextLabel();
	l->at = 0;		/* Where is the jump */
	l->to = offset;		/* What place does it goto */
	l->from = 0;
	l->type = Lcode|Linternal;
	return (l);
}

label*
reference_table_label(int32 n)
{
	label* l;

	assert(n < MAXLABTAB);
	if (labtab[n] == 0) {
		l = nextLabel();
		labtab[n] = l;
		l->type = Lnull;
		l->at = 0;
		l->from = 0;
		l->to = 0;
	}
	else {
		l = labtab[n];
		labtab[n] = 0;
	}
	return (l);
}

SlotInfo*
stored_code_label(SlotInfo* dst)
{
	return (dst);
}

SlotInfo*
table_code_label(SlotInfo* dst)
{
	return (dst);
}

#if defined(HAVE_set_label)
void
set_label(int i, int n)
{
	assert(n < MAXLABTAB);
	if (labtab[n] == 0) {
		labtab[n] = nextLabel();
		labtab[n]->type = Lgeneral|Linternal;
		labtab[n]->at = 0;
		labtab[n]->from = 0;
		labtab[n]->to = 0;
		CALL_GEN_FUNC(set_label)(NULLSLOT, NULLSLOT, (int)labtab[n]);
	}
	else {
		assert((labtab[n]->type & (Lgeneral|Linternal)) == Lnull);
		labtab[n]->type |= Lgeneral|Linternal;
		CALL_GEN_FUNC(set_label)(NULLSLOT, NULLSLOT, (int)labtab[n]);
		labtab[n] = 0;
	}
}
#endif

#if defined(HAVE_build_code_ref)
label*
build_code_ref(uint8* pos, uintp pc)
{
	label* l;
	int offset;

	offset = (int32)(pos[0] * 0x01000000 + pos[1] * 0x00010000 +
		pos[2] * 0x00000100 + pos[3] * 0x00000001);
	l = reference_code_label(pc+offset);

	CALL_GEN_FUNC(build_code_ref)(NULLSLOT, NULLSLOT, (int)l);
	return (l);
}
#endif

#if defined(HAVE_build_key)
void
build_key(uint8* pos)
{
	jint val = pos[0] * 0x01000000 + pos[1] * 0x00010000 +
		pos[2] * 0x00000100 + pos[3] * 0x00000001;

	CALL_GEN_FUNC(build_key)(NULLSLOT, NULLSLOT, val);
}
#endif


/* ----------------------------------------------------------------------- */
/* Comparisons.								   */
/*									   */

void
cmp_int_const(SlotInfo* dst, SlotInfo* src, jint val)
{
#if defined(HAVE_cmp_int_const)
	CALL_GEN_FUNC(cmp_int_const)(dst, src, val);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);
	move_int_const(tmp, val);
	cmp_int(dst, src, tmp);
	slot_freetmp(tmp);
#endif
}

#if defined(HAVE_cmp_int)
void
cmp_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_cmp_int_const)
	if (rslot_isconst(src2)) {
		cmp_int_const(dst, src, rslot_value(src2));
		return;
	}
#endif
	CALL_GEN_FUNC(cmp_int)(dst, src, src2);
}
#endif

void
cmp_load_offset_int(SlotInfo* dst, SlotInfo* src, SlotInfo* src2, int offset)
{
#if defined(HAVE_cmp_load_offset_int)
	CALL_GEN_FUNC(cmp_load_offset_int)(dst, src, src2, offset);
#else
	SlotInfo* tmp;
	slot_alloctmp(tmp);

	load_int_offset(tmp, src2, offset);
	cmp_int(dst, src, tmp);

	slot_freetmp(tmp);
#endif
}

void
cmp_long(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_cmp_long)
	CALL_GEN_FUNC(cmp_long)(dst, src, src2);
#else
	cbranch_int(src+1, src2+1, reference_label(0, 1), bgt);
	branch(reference_label(0, 2), bne);

	cbranch_int(src, src2, reference_label(0, 3), beq);
	branch(reference_label(0, 4), bult);

	set_label(0, 1);
	move_int_const(dst, -1);
	branch(reference_label(0, 5), ba);

	set_label(0, 3);
	move_int_const(dst, 0);
	branch(reference_label(0, 6), ba);

	set_label(0, 2);
	set_label(0, 4);
	move_int_const(dst, 1);

	set_label(0, 5);
	set_label(0, 6);
#endif
#if 0
	slot_nowriteback2(src2);
	slot_nowriteback2(src);
	prepare_function_call();
	pusharg_long(src2);
	pusharg_long(src);
	call_ref((uintp)soft_lcmp);
	popargs();
	fixup_function_call();
	return_int(dst);
	reload_inner();
#endif
}

void
cmpl_float(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_cmpl_float)
	abort();
#else
	prepare_function_call();
	pusharg_float(src2);
	pusharg_float(src);
	call_ref((uintp)soft_fcmpl);
	popargs();
	fixup_function_call();
	return_int(dst);
	reload_inner();
#endif
}

void
cmpl_double(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_cmpl_double)
	abort();
#else
	prepare_function_call();
	pusharg_double(src2);
	pusharg_double(src);
	call_ref((uintp)soft_dcmpl);
	popargs();
	fixup_function_call();
	return_int(dst);
	reload_inner();
#endif
}

void
cmpg_float(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_cmpg_float)
	abort();
#else
	prepare_function_call();
	pusharg_float(src2);
	pusharg_float(src);
	call_ref((uintp)soft_fcmpg);
	popargs();
	fixup_function_call();
	return_int(dst);
	reload_inner();
#endif
}

void
cmpg_double(SlotInfo* dst, SlotInfo* src, SlotInfo* src2)
{
#if defined(HAVE_cmpg_double)
	abort();
#else
	prepare_function_call();
	pusharg_double(src2);
	pusharg_double(src);
	call_ref((uintp)soft_dcmpg);
	popargs();
	fixup_function_call();
	return_int(dst);
	reload_inner();
#endif
}

/* ----------------------------------------------------------------------- */
/* Conversions.								   */
/*									   */

void
cvt_int_long(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_cvt_int_long) 
	CALL_GEN_FUNC(cvt_int_long)(dst, NULLSLOT, src);
#else
	move_int(dst, src);
	ashr_int_const(dst+1, dst, (8 * sizeof(jint)) - 1);
#endif
}

#if defined(HAVE_cvt_int_float) 
void
cvt_int_float(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_int_float)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_cvt_int_double) 
void
cvt_int_double(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_int_double)(dst, NULLSLOT, src);
}
#endif

void
cvt_long_int(SlotInfo* dst, SlotInfo* src)
{
	move_int(dst, src);
}

#if defined(HAVE_cvt_long_float) 
void
cvt_long_float(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_long_float)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_cvt_long_double) 
void
cvt_long_double(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_long_double)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_cvt_float_int) 
void
cvt_float_int(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_float_int)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_cvt_float_long) 
void
cvt_float_long(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_float_long)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_cvt_float_double) 
void
cvt_float_double(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_float_double)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_cvt_double_int) 
void
cvt_double_int(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_double_int)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_cvt_double_long) 
void
cvt_double_long(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_double_long)(dst, NULLSLOT, src);
}
#endif

#if defined(HAVE_cvt_double_float) 
void
cvt_double_float(SlotInfo* dst, SlotInfo* src)
{
	CALL_GEN_FUNC(cvt_double_float)(dst, NULLSLOT, src);
}
#endif

void
cvt_int_byte(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_cvt_int_byte) 
	CALL_GEN_FUNC(cvt_int_byte)(dst, NULLSLOT, src);
#else
	lshl_int_const(dst, src, 8 * (sizeof(jint) - sizeof(jbyte)));
	ashr_int_const(dst, dst, 8 * (sizeof(jint) - sizeof(jbyte)));
#endif
}

void
cvt_int_char(SlotInfo* dst, SlotInfo* src)
{
	and_int_const(dst, src, (1 << (8 * sizeof(jchar))) - 1);
}

void
cvt_int_short(SlotInfo* dst, SlotInfo* src)
{
#if defined(HAVE_cvt_int_short) 
	CALL_GEN_FUNC(cvt_int_short)(dst, NULLSLOT, src);
#else
	lshl_int_const(dst, src, 8 * (sizeof(jint) - sizeof(jshort)));
	ashr_int_const(dst, dst, 8 * (sizeof(jint) - sizeof(jshort)));
#endif
}

void
boundscheck(SlotInfo* obj, SlotInfo* idx)
{
#if defined(HAVE_boundscheck)
	abort();
#else
	cbranch_load_offset_int_ult(idx, obj, object_array_length, reference_label(0, 10));
	softcall_badarrayindex();
	set_label(0, 10);
#endif
}

void
softcall_checkarraystore(SlotInfo* array, SlotInfo* obj)
{
	prepare_function_call();
	pusharg_ref(obj);
	pusharg_ref(array);  
	call_ref((uintp)soft_checkarraystore);
	popargs();
	fixup_function_call();
}

void
checknull(SlotInfo* obj)
{
	cbranch_ref_const_ne(obj, 0, reference_label(0, 34));
	softcall_nullpointer();
	set_label(0, 34);
}


/* ----------------------------------------------------------------------- */
/* Breakpoints.								   */
/*									   */

void
breakpoint()
{
	abort();
}


/* ----------------------------------------------------------------------- */
/* Soft calls.								   */
/*									   */

void
softcall_lookupmethod(SlotInfo* dst, Method* meth, SlotInfo* obj)
{
	/* 'obj' must be written back since it will be reused by the call */
	prepare_function_call();
	pusharg_ref_const(meth);
	pusharg_ref(obj);
	call_ref((uintp)soft_lookupmethod);
	popargs();
	fixup_function_call();
	return_ref(dst);
	reload_inner();
}

void
softcall_get_method_code (SlotInfo* dst, SlotInfo* method)
{
	slot_nowriteback(method);
	prepare_function_call();
	pusharg_ref(method);
	call_ref((uintp)soft_get_method_code);
	popargs();
	fixup_function_call();
	return_ref(dst);
}

void
softcall_get_method_code_const (SlotInfo* dst, Method* method)
{
	if (method->ncode == 0) {
		prepare_function_call();
		pusharg_ref_const(method);
		call_ref((uintp)soft_get_method_code);
		popargs();
		fixup_function_call();
		return_ref(dst);
	}
	else {
		move_ref_const(dst, method->ncode);
	}
}

void
softcall_badarrayindex(void)
{
	call_ref((uintp)soft_badarrayindex);
}

void
softcall_nullpointer(void)
{
	call_ref((uintp)soft_nullpointer);
}

void
softcall_new(SlotInfo* dst, classes* classobj)
{
	prepare_function_call();
	pusharg_ref_const(classobj);
	call_ref((uintp)soft_new);
	popargs();
	fixup_function_call();
	return_ref(dst);
	reload_inner();
}

void
softcall_newarray(SlotInfo* dst, SlotInfo* size, int type)
{
	slot_nowriteback(size);
	prepare_function_call();
	pusharg_int(size);
	pusharg_int_const(type);
	call_ref((uintp)soft_newarray);
	popargs();
	fixup_function_call();
	return_ref(dst);
	reload_inner();
}

void
softcall_anewarray(SlotInfo* dst, SlotInfo* size, classes* type)
{
	slot_nowriteback(size);
	prepare_function_call();
	pusharg_int(size);
	pusharg_ref_const(type);
	call_ref((uintp)soft_anewarray);
	popargs();
	fixup_function_call();
	return_ref(dst);
	reload_inner();
}

void
softcall_multianewarray(SlotInfo* dst, int size, SlotInfo* stktop, classes* classobj)
{
	int i;

	prepare_function_call();
	for (i = 0; i < size; i++) {
		pusharg_ref(&stktop[i]);
	}
	pusharg_int_const(size);
	pusharg_ref_const(classobj);
	call_ref((uintp)soft_multianewarray);
	popargs();
	fixup_function_call();
	return_ref(dst);
	reload_inner();
}

void
softcall_athrow(SlotInfo* obj)
{
	slot_nowriteback(obj);
	prepare_function_call();
	pusharg_ref(obj);
	call_ref((uintp)soft_athrow);
	popargs();
	fixup_function_call();
	reload_inner();
}

void
softcall_checkcast(SlotInfo* res, SlotInfo* obj, classes* class)
{
	prepare_function_call();
	pusharg_ref(obj);
	pusharg_ref_const(class);
	call_ref((uintp)soft_checkcast);
	popargs();
	fixup_function_call();
	reload_inner();
	move_ref(res, obj);
}

void
softcall_instanceof(SlotInfo* dst, SlotInfo* obj, classes* class)
{
	slot_nowriteback(obj);
	prepare_function_call();
	pusharg_ref(obj);
	pusharg_ref_const(class);
	call_ref((uintp)soft_instanceof);
	popargs();
	fixup_function_call();
	return_int(dst);
	reload_inner();
}

void
softcall_monitorenter(SlotInfo* mon)
{
	slot_nowriteback(mon);
	prepare_function_call();
	pusharg_ref(mon);
	call_ref((uintp)soft_monitorenter);
	popargs();
	fixup_function_call();
	reload_inner();
}

void
softcall_monitorexit(SlotInfo* mon)
{
	slot_nowriteback(mon);
	prepare_function_call();
	pusharg_ref(mon);
	call_ref((uintp)soft_monitorexit);
	popargs();
	fixup_function_call();
	reload_inner();
}

void
softcall_initialise_class(classes* c)
{
	SlotInfo* tmp2;

	slot_alloctmp(tmp2);
	load_addr_int(tmp2, (uintp)c + CLASS_STATE);
	end_basic_block();
	cbranch_int_const_eq(tmp2, CSTATE_OK, reference_label(0, 38));
	prepare_function_call();
	pusharg_ref_const(c);
	call_ref((uintp)soft_initialise_class);
	popargs();
	fixup_function_call();
	reload_inner();
	set_label(0, 38);
	slot_freetmp(tmp2);
	start_basic_block();
}

#if defined(GC_INCREMENTAL)
void
softcall_addreference(SlotInfo* from, SlotInfo* to)
{
	prepare_function_call();
	pusharg_ref(to);
	pusharg_ref(from);
	call_ref((uintp)soft_addreference);
	popargs();
	fixup_function_call();
}

void
softcall_addreference_static(void* from, SlotInfo* to)
{
	prepare_function_call();
	pusharg_ref(to);
	pusharg_ref_const(from);
	call_ref((uintp)soft_addreference);
	popargs();
	fixup_function_call();
}
#endif
