/* labels.c
 * Manage the code labels and links.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "labels.h"
#include "code-analyse.h"

label* firstLabel;
label* lastLabel;
label* currLabel;

void
initLabels(void)
{
	currLabel = firstLabel;
}

void
linkLabels(void)
{
	uintp dest;
	uintp from;
	uintp* place;
	label* l;

	for (l = firstLabel; l != currLabel; l = l->next) {

		/* Ignore cancelled labels */
		if (l->type == Lcancelled) {
			continue;
		}

DBG(		printf("label 0x%x: %x %x %x\n", l, l->at, l->to, l->from); )

		if ((l->type & Llong) != Llong) {
			printf("Label not long (%x).\n", (uintp)l);
			abort();
		}

		/* Find destination of jump */
		if (l->type & Lcode) {
			assert(codeInfo.perPC[l->to].pc != (uintp)-1);
			dest = codeInfo.perPC[l->to].pc;
		}
		else {
			dest = l->to;
		}

		/* Find place to store result */
		place = (uintp*)(l->at);

		/* If label is relative, find where it is relative to */
		if (l->type & Lrelative) {
			from = l->from;
		}
		else {
			from = 0;
		}

		/* Install result */
		(*place) = dest - from;
	}
}

/*
 * Allocate a new label.
 */
label*
nextLabel(void)
{
	int i;
	label* ret = currLabel;

	if (ret == 0) {
		/* Allocate chunk of label elements */
		ret = calloc(ALLOCLABELNR, sizeof(label));
		assert(ret != 0);

		/* Attach to current chain */
		if (lastLabel == 0) {
			firstLabel = ret;
		}
		else {
			lastLabel->next = ret;
		}
		lastLabel = &ret[ALLOCLABELNR-1];

		/* Link elements into list */
		for (i = 0; i < ALLOCLABELNR-1; i++) {
			ret[i].next = &ret[i+1];
		}
		ret[ALLOCLABELNR-1].next = 0;
	}
	currLabel = ret->next;
	return (ret);
}
