/* codeslots.h
 * Code slots.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#ifndef __codeslots_h
#define __codeslots_h

#include "gtypes.h"

#define	MAXSLOT			64

#define Tcomplex		0
#define Tnull			Tcomplex
#define Tconst			1
#define Tcopy			2
#define Taddregconst		3
#define Tstore			4
#define Tload			5

struct _SlotInfo;
struct _slotreg;

typedef struct _typedata {
	struct _slotreg*	free;
	struct _slotreg*	inuse;
} typedata;

typedef struct _SlotInfo {
	uint8			consttype;
        union {
                jint            tint;
                jlong           tlong;
                jdouble         tdouble;
                void*           taddr;
                char*           tstr;
        } v;
	int16			type;
	struct _slotreg*	reg;
	struct _slotreg*	fixedreg;
	int32			offset;
	struct _SlotInfo*	next;
} SlotInfo;

typedef struct _slotreg {
	int16			type;
	int16			ctype;
	struct _slotreg*	next;
	struct _SlotInfo*	myslot;
	int32			issuenr;
	int8			regno;
	int8			modified;
} slotreg;

#define	MAX_TRANSLATION_TYPES		9

#define Rint				0
#define Rlong				1
#define Rfloat				2
#define Rdouble				3
#define Rref				4
#define	Rvoid				5
#define	Raddr				6
#define	Rany				7
#define	Ranylong			8

extern SlotInfo* cslotinfo;
extern SlotInfo* localinfo;
extern SlotInfo* stackinfo;
extern SlotInfo* tempinfo;
extern SlotInfo* returninfo;
extern int tmpslot;
extern slotreg registerData[];

#define	REGDATA(r)	&registerData[r]

struct _innerLoop;
void initCodeSlots(int);
void freeSlot(SlotInfo*);
void forceRegister(SlotInfo*, slotreg*, int, uint32);
void changeRegister(SlotInfo*, slotreg*, int, uint32);
void clobberRegister(slotreg*);
int demandRegister(SlotInfo*, int, int, uint32);
void activateInner(struct _innerLoop*, int);

/*---------------------------------------------------------------------*/

#define	stack(_s)		(&localinfo[stackno+(_s)])
#define	rstack(_s)		(&localinfo[ISLOTS(pc, stackno+(_s))])
#define	wstack(_s)		(&localinfo[OSLOTS(pc, stackno+(_s))])
#define	rstack_value(_s)	(IVALUES(pc, stackno+(_s)))
#define	rstack_isconst(_s)	(ITYPES(pc, stackno+(_s)) == SCONST)
#define	local(_s)		(&localinfo[(_s)])
#define	local_long(_s)		(&localinfo[(_s)])
#define	local_float(_s)		local(_s)
#define	local_double(_s)	local_long(_s)

/*---------------------------------------------------------------------*/

#define	slot_return(_s)		((_s) >= &returninfo[0] &&		\
				 (_s) < &returninfo[MAXRETURNS])
#define	slot_local(_s)		((_s) >= &localinfo[0] &&		\
				 (_s) < &localinfo[maxLocal])
#define	slot_stack(_s)		((_s) >= &stackinfo[0] &&		\
				 (_s) < &stackinfo[maxStack])
#define	slot_tmp(_s)		((_s) >= &tempinfo[0])

/* Does this slot have a register allocated to it? */
#define	register_is_allocated(_s)	((_s)->reg != 0)

#define	rslot_isconst(_s)	(slot_stack(_s) && \
					ITYPES(pc,((_s)-localinfo)) == SCONST)
#define	rslot_value(_s)		(IVALUES(pc, ((_s)-localinfo)))

#define	slot_alloctmp(t)					\
		(t) = &tempinfo[tmpslot],			\
		tmpslot += 1
#define	slot_alloc2tmp(t)					\
		(t) = &tempinfo[tmpslot],			\
		tmpslot += 2
#define	slot_freetmp(t)						\
		freeSlot((t))
#define	slot_free2tmp(t)					\
		freeSlot((t));					\
		freeSlot((t)+1)

#define	slot_nowriteback(_s)					\
		if ((_s)->reg != 0 && slot_stack(_s)) {		\
			(_s)->reg->modified |= rnowriteback;	\
		}

#define	slot_nowriteback2(_s)					\
		slot_nowriteback(_s);				\
		slot_nowriteback((_s)+1)

/* Return slots */      
#define	MAXRETURNS		7
#define	returnInt		0
#define	returnRef		1
#define	returnLong		2 /* & 3 */
#define	returnFloat		4
#define	returnDouble		5 /* & 6 */

/* Register flags */
#define rwrite			1
#define rread			2
#define rnowriteback		4

#define	rreserved		8
#define	rnocallsave		16

#define SLOT2FRAMEOFFSET(_s)					\
		(((_s) - cslotinfo < maxArgs) ?			\
			SLOT2ARGOFFSET((_s) - cslotinfo) :	\
			SLOT2LOCALOFFSET((_s) - cslotinfo))

/*---------------------------------------------------------------------*/

#if defined(WRITEBACK_INTEGER)
#define	WBI	rwrite
#else
#define	WBI	0
#endif
#if defined(WRITEBACK_LONG)
#define	WBL	rwrite
#else
#define	WBL	0
#endif
#if defined(WRITEBACK_FLOAT)
#define	WBF	rwrite
#else
#define	WBF	0
#endif
#if defined(WRITEBACK_DOUBLE)
#define	WBD	rwrite
#else
#define	WBD	0
#endif

#define	const_int(_i)		(*(jint*)&s##_i)
#define	const_long(_i)		(*(jlong*)&s##_i)
#define	const_float(_i)		(*(jfloat*)&s##_i)
#define	const_double(_i)	(*(jdouble*)&s##_i)

#define	sreg_int(_i)		((s##_i)->reg->regno)
#define	sreg_long(_i)		((s##_i)->reg->regno)
#define	sreg_float(_i)		((s##_i)->reg->regno)
#define	sreg_double(_i)		((s##_i)->reg->regno)

#define	lreg_int(_i)		((s##_i)->reg->regno)
#define	lreg_long(_i)		((s##_i)->reg->regno)
#define	lreg_float(_i)		((s##_i)->reg->regno)
#define	lreg_double(_i)		((s##_i)->reg->regno)

#define	rreg_int(_i)		demandRegister(s##_i, rread|WBI, Rint, issuenr)
#define	wreg_int(_i)		demandRegister(s##_i, rwrite, Rint, issuenr)
#define	rwreg_int(_i)		demandRegister(s##_i, rread|rwrite, Rint, issuenr)
#define	rslot_int(_i)		demandOffset(s##_i, rread)
#define	wslot_int(_i)		demandOffset(s##_i, rwrite)

#define	rreg_long(_i)		demandRegister(s##_i, rread|WBL, Rlong, issuenr)
#define	wreg_long(_i)		demandRegister(s##_i, rwrite, Rlong, issuenr)
#define	rwreg_long(_i)		demandRegister(s##_i, rread|rwrite, Rlong, issuenr)
#define	rslot_long(_i)		demandOffset(s##_i, rread); \
				demandOffset(1+s##_i, rread)
#define	wslot_long(_i)		demandOffset(s##_i, rwrite); \
				demandOffset(1+s##_i, rwrite)

#define	rreg_float(_i)		demandRegister(s##_i, rread|WBF, Rfloat, issuenr)
#define	wreg_float(_i)		demandRegister(s##_i, rwrite, Rfloat, issuenr)
#define	rwreg_float(_i)		demandRegister(s##_i, rread|rwrite, Rfloat, issuenr)
#define	rslot_float(_i)		demandOffset(s##_i, rread)
#define	wslot_float(_i)		demandOffset(s##_i, rwrite)

#define	rreg_double(_i)		demandRegister(s##_i, rread|WBD, Rdouble, issuenr)
#define	wreg_double(_i)		demandRegister(s##_i, rwrite, Rdouble, issuenr)
#define	rwreg_double(_i)	demandRegister(s##_i, rread|rwrite, Rdouble, issuenr)
#define	rslot_double(_i)	demandOffset(s##_i, rread); \
				demandOffset(1+s##_i, rread)
#define	wslot_double(_i)	demandOffset(s##_i, rwrite); \
				demandOffset(1+s##_i, rwrite)

#endif
